import logging
from collections import defaultdict
from enum import Enum
from typing import Any, Dict

import inflection

from protean.utils import DomainObjects, fully_qualified_name

logger = logging.getLogger(__name__)


# Define property names for each element type
def properties() -> Dict[str, str]:
    """Properties are named after of each element type and pluralized to
    indicate that all elements of the type will be returned.

    E.g.
    AGGREGATE: registry.aggregates
    VALUE_OBJECT: registry.value_objects

    Returns:
        Dict[str, str]: Dict[pluralized_name, element_type], a dictionary of element type names and their values.
        E.g.
            {
                'aggregates': 'AGGREGATE',
                'entities': 'ENTITY',
                ...
            }
    """
    props = {}
    for element_type in DomainObjects:
        # Lowercase element type, add underscores and pluralize
        prop_name = inflection.pluralize(
            inflection.underscore(element_type.value.lower())
        )
        props[prop_name] = element_type.value

    return props


class _DomainRegistry:
    class DomainRecord:
        def __init__(self, name: str, qualname: str, class_type: str, cls: Any):
            self.name = name
            self.qualname = qualname
            self.class_type = class_type
            self.cls = cls

        def __repr__(self):
            return f"<class {self.name}: {self.qualname} ({self.class_type})>"

    def __init__(self):
        self._elements: Dict[str, dict] = {}
        self._elements_by_name: Dict[str, list] = {}

        # Initialize placeholders for element types
        for element_type in DomainObjects:
            self._elements[element_type.value] = defaultdict(dict)

    def _reset(self):
        for element_type in DomainObjects:
            self._elements[element_type.value] = defaultdict(dict)
        self._elements_by_name: Dict[str, list] = {}

    def _is_invalid_element_cls(self, element_cls):
        """Ensure that we are dealing with an element class, that:

        * Has a `element_type` attribute
        * `element_type` is an Enum value
        * The value of `element_type` enum is among recognized `DomainObjects` values
        """
        return (
            not hasattr(element_cls, "element_type")
            or not isinstance(element_cls.element_type, Enum)
            or element_cls.element_type.name not in DomainObjects.__members__
        )

    def register_element(self, element_cls):
        if self._is_invalid_element_cls(element_cls):
            raise NotImplementedError

        element_name = fully_qualified_name(element_cls)

        element = self._elements[element_cls.element_type.value][element_name]
        if element:
            # raise ConfigurationError(f'Element {element_name} has already been registered')
            logger.debug(f"Element {element_name} was already in the registry")
        else:
            element_record = _DomainRegistry.DomainRecord(
                name=element_cls.__name__,
                qualname=element_name,
                class_type=element_cls.element_type.value,
                cls=element_cls,
            )

            self._elements[element_cls.element_type.value][element_name] = (
                element_record
            )

            # Create an array to hold multiple elements of same name
            if element_cls.__name__ in self._elements_by_name:
                self._elements_by_name[element_cls.__name__].append(element_record)
            else:
                self._elements_by_name[element_cls.__name__] = [element_record]

            logger.debug(
                f"Registered Element {element_name} with Domain as a {element_cls.element_type.value}"
            )

    def delist_element(self, element_cls):
        if self._is_invalid_element_cls(element_cls):
            raise NotImplementedError

        element_name = fully_qualified_name(element_cls)

        self._elements[element_cls.element_type.value].pop(element_name, None)

    @property
    def elements(self):
        elems = {}
        for name, element_type in properties().items():
            items = []
            for item in self._elements[element_type]:
                # Add only the class of the element
                items.append(self._elements[element_type][item].cls)

            if items:  # Only add element type if there are elements of that type
                elems[name] = items

        return elems

    def __repr__(self):
        return f"<DomainRegistry: {self._elements}>"


# Set up access to all elements as properties
for name, element_type in properties().items():
    """Set up element types as properties on Registry for easy access.

    Why? Since all elements are stored within a Dict in the registry, accessing
    them will mean knowing the storage structure. It is instead preferable to
    expose the elements by their element types as properties.
    """

    # This weird syntax is because when using lambdas in a for loop, we need to supply
    #   element_type as an argument with a default value of element_type
    prop = property(
        lambda self, element_type=element_type: self._elements[element_type]
    )  # pragma: no cover  # FIXME Is it possible to cover this line in tests

    # Set the property on the class
    setattr(_DomainRegistry, name, prop)



"""Value Object Functionality and Classes"""

import inspect
import logging
from collections import defaultdict

from protean.container import BaseContainer, OptionsMixin, fields
from protean.exceptions import IncorrectUsageError, NotSupportedError, ValidationError
from protean.fields import Reference, ValueObject
from protean.fields.association import Association
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger(__name__)


class BaseValueObject(BaseContainer, OptionsMixin):
    element_type = DomainObjects.VALUE_OBJECT

    def __new__(cls, *args, **kwargs):
        if cls is BaseValueObject:
            raise NotSupportedError("BaseValueObject cannot be instantiated")
        return super().__new__(cls)

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        # Record invariant methods
        setattr(subclass, "_invariants", defaultdict(dict))

        subclass.__validate_for_basic_field_types()
        subclass.__validate_for_non_identifier_fields()
        subclass.__validate_for_non_unique_fields()

    @classmethod
    def __validate_for_basic_field_types(subclass):
        for field_name, field_obj in fields(subclass).items():
            if isinstance(field_obj, (Reference, Association, ValueObject)):
                raise IncorrectUsageError(
                    {
                        "_value_object": [
                            f"Value Objects can only contain basic field types. "
                            f"Remove {field_name} ({field_obj.__class__.__name__}) from class {subclass.__name__}"
                        ]
                    }
                )

    @classmethod
    def __validate_for_non_identifier_fields(subclass):
        for field_name, field_obj in fields(subclass).items():
            if field_obj.identifier:
                raise IncorrectUsageError(
                    {
                        "_value_object": [
                            f"Value Objects cannot contain fields marked 'identifier' (field '{field_name}')"
                        ]
                    }
                )

    @classmethod
    def __validate_for_non_unique_fields(subclass):
        for field_name, field_obj in fields(subclass).items():
            if field_obj.unique:
                raise IncorrectUsageError(
                    {
                        "_value_object": [
                            f"Value Objects cannot contain fields marked 'unique' (field '{field_name}')"
                        ]
                    }
                )

    def __init__(self, *template, **kwargs):  # noqa: C901
        """
        Initialise the container.

        During initialization, set value on fields if validation passes.

        This initialization technique supports keyword arguments as well as dictionaries. You
            can even use a template for initial data.

        FIXME Can we depend on Container's Init and only implement VO specific aspects here?
        """

        if self.meta_.abstract is True:
            raise NotSupportedError(
                f"{self.__class__.__name__} class has been marked abstract"
                f" and cannot be instantiated"
            )

        self.errors = defaultdict(list)

        # Set the flag to prevent any further modifications
        self._initialized = False

        # Load the attributes based on the template
        loaded_fields = []
        for dictionary in template:
            if not isinstance(dictionary, dict):
                raise AssertionError(
                    f"Positional argument {dictionary} passed must be a dict. "
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in dictionary.items():
                loaded_fields.append(field_name)
                setattr(self, field_name, val)

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            # Record that a field was encountered by appending to `loaded_fields`
            #   When it fails validations, we want it's errors to be recorded
            #
            #   Not remembering the field was recorded will result in it being set to `None`
            #   which will raise a ValidationError of its own for the wrong reasons (required field not set)
            loaded_fields.append(field_name)
            try:
                setattr(self, field_name, val)
            except ValidationError as err:
                for field_name in err.messages:
                    self.errors[field_name].extend(err.messages[field_name])

        # Now load the remaining fields with a None value, which will fail
        # for required fields
        for field_name in fields(self):
            if field_name not in loaded_fields:
                setattr(self, field_name, None)

        self.defaults()

        # `_postcheck()` will return a `defaultdict(list)` if errors are to be raised
        custom_errors = self._postcheck() or {}
        for field in custom_errors:
            self.errors[field].extend(custom_errors[field])

        # Raise any errors found during load
        if self.errors:
            logger.error(self.errors)
            raise ValidationError(self.errors)

        # If we made it this far, the Value Object is initialized
        #   and should be marked as such
        self._initialized = True

    def __setattr__(self, name, value):
        if not hasattr(self, "_initialized") or not self._initialized:
            return super().__setattr__(name, value)
        else:
            raise IncorrectUsageError(
                {
                    "_value_object": [
                        "Value Objects are immutable and cannot be modified once created"
                    ]
                }
            )

    def _postcheck(self, return_errors=False):
        """Invariant checks performed after initialization"""
        errors = defaultdict(list)

        for invariant_method in self._invariants["post"].values():
            try:
                invariant_method(self)
            except ValidationError as err:
                for field_name in err.messages:
                    errors[field_name].extend(err.messages[field_name])

        if return_errors:
            return errors

        if errors:
            raise ValidationError(errors)


def value_object_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseValueObject, **kwargs)

    # Iterate through methods marked as `@invariant` and record them for later use
    methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
    for method_name, method in methods:
        if not (
            method_name.startswith("__") and method_name.endswith("__")
        ) and hasattr(method, "_invariant"):
            element_cls._invariants[method._invariant][method_name] = method

    return element_cls



"""This module implements the central domain object, along with decorators
to register Domain Elements.
"""

import inspect
import logging
import sys
from collections import defaultdict
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from werkzeug.datastructures import ImmutableDict

from protean.adapters import Brokers, Caches, EmailProviders, Providers
from protean.adapters.event_store import EventStore
from protean.core.command import BaseCommand
from protean.core.command_handler import BaseCommandHandler
from protean.core.event import BaseEvent
from protean.core.event_handler import BaseEventHandler
from protean.core.model import BaseModel
from protean.core.repository import BaseRepository
from protean.domain.registry import _DomainRegistry
from protean.exceptions import ConfigurationError, IncorrectUsageError
from protean.fields import HasMany, HasOne, Reference, ValueObject
from protean.reflection import declared_fields, has_fields
from protean.utils import (
    CommandProcessing,
    DomainObjects,
    EventProcessing,
    fqn,
)

from .config import Config2, ConfigAttribute
from .context import DomainContext, _DomainContextGlobals
from .helpers import get_debug_flag, get_env

logger = logging.getLogger(__name__)

# a singleton sentinel value for parameter defaults
_sentinel = object()


class Domain:
    """The domain object is a one-stop gateway to:

    * Registering Domain Objects/Concepts
    * Querying/Retrieving Domain Artifacts like Entities, Services, etc.
    * Retrieve injected infrastructure adapters

    Usually you create a :class:`Domain` instance in your main module or
    in the :file:`__init__.py` file of your package like this::

        from protean import Domain
        domain = Domain(__name__)

    :param domain_name: the name of the domain
    """

    from protean.utils import IdentityStrategy, IdentityType

    config_class = Config2
    domain_context_globals_class = _DomainContextGlobals

    #: What environment the app is running in Protean and extensions may
    #: enable behaviors based on the environment, such as enabling debug
    #: mode. This maps to the :data:`ENV` config key. This is set by the
    #: :envvar:`PROTEAN_ENV` environment variable and may not behave as
    #: expected if set in code.
    #:
    #: **Do not enable development when deploying in production.**
    #:
    #: Default: ``'production'``
    env = ConfigAttribute("env")

    #: The testing flag.  Set this to ``True`` to enable the test mode of
    #: Protean extensions (and in the future probably also Protean itself).
    #: For example this might activate test helpers that have an
    #: additional runtime cost which should not be enabled by default.
    #:
    #: This attribute can also be configured from the config with the
    #: ``testing`` configuration key.  Defaults to ``False``.
    testing = ConfigAttribute("testing")

    #: If a secret key is set, cryptographic components can use this to
    #: sign cookies and other things. Set this to a complex random value
    #: when you want to use the secure cookie for instance.
    #:
    #: This attribute can also be configured from the config with the
    #: :data:`secret_key` configuration key. Defaults to ``None``.
    secret_key = ConfigAttribute("secret_key")

    default_config = ImmutableDict(
        {
            "env": None,
            "debug": None,
            "secret_key": None,
            "identity_strategy": IdentityStrategy.UUID.value,
            "identity_type": IdentityType.STRING.value,
            "databases": {
                "default": {"provider": "memory"},
                "memory": {"provider": "memory"},
            },
            "event_processing": EventProcessing.ASYNC.value,
            "command_processing": CommandProcessing.ASYNC.value,
            "event_store": {
                "provider": "memory",
            },
            "caches": {
                "default": {
                    "provider": "memory",
                    "TTL": 300,
                }
            },
            "brokers": {"default": {"provider": "inline"}},
            "EMAIL_PROVIDERS": {
                "default": {
                    "provider": "protean.adapters.DummyEmailProvider",
                    "DEFAULT_FROM_EMAIL": "admin@team8solutions.com",
                },
            },
            "SNAPSHOT_THRESHOLD": 10,
        },
    )

    def __init__(
        self,
        root_path: str,
        name: str = "",
        load_toml: bool = True,
        identity_function: Optional[Callable] = None,
    ):
        self.root_path = root_path

        # Initialize the domain with the name of the module if not provided
        # Get the stack frame of the caller of the __init__ method
        caller_frame = inspect.stack()[1]
        # Get the module name from the globals of the frame where the object was instantiated
        self.name = name if name else caller_frame.frame.f_globals["__name__"]

        # FIXME Additional domain attributes: (Think if this is needed)
        #   - Type of Domain: Core, Supporting, Third-party(?)
        #   - Type of Implementation: CRUD, CQRS, ES

        # Registry for all domain Objects
        self._domain_registry = _DomainRegistry()

        #: The configuration dictionary as :class:`Config`.  This behaves
        #: exactly like a regular dictionary but supports additional methods
        #: to load a config from files.
        self.config = self.load_config(load_toml)

        # The function to invoke to generate identity
        self._identity_function = identity_function

        self.providers = Providers(self)
        self.event_store = EventStore(self)
        self.brokers = Brokers(self)
        self.caches = Caches(self)
        self.email_providers = EmailProviders(self)

        # Cache for holding Model to Entity/Aggregate associations
        self._models: Dict[str, BaseModel] = {}
        self._constructed_models: Dict[str, BaseModel] = {}

        #: A list of functions that are called when the domain context
        #: is destroyed.  This is the place to store code that cleans up and
        #: disconnects from databases, for example.
        self.teardown_domain_context_functions: List[Callable] = []

        # Placeholder array for resolving classes referenced by domain elements
        # FIXME Should all protean elements be subclassed from a base element?
        self._pending_class_resolutions: dict[str, Any] = defaultdict(list)

        # Initialize domain with default adapters
        self._initialize()

    def init(self, traverse=True):  # noqa: C901
        """Parse the domain folder, and attach elements dynamically to the domain.

        Protean parses all files in the domain file's folder, as well as under it,
        to load elements. So, all domain files are to be nested under the file contain
        the domain definition.

        One can use the `traverse` flag to control this functionality, `True` by default.

        When enabled, Protean is responsible for loading domain elements and ensuring
        all functionality is activated.

        The developer is responsible for activating functionality manually when
        autoloading is disabled. Element activation can be done by importing them
        in central areas of domain execution, like Application Services.

        For example, asynchronous aspects of a domain like its Subscribers and
        Event Handlers should be imported in their relevant Application Services
        and Aggregates.

        This method bubbles up circular import issues, if present, in the domain code.
        """
        if traverse is True:
            self._traverse()

        # Resolve all pending references
        self._resolve_references()

        # Assign Aggregate Clusters
        self._assign_aggregate_clusters()

        # Run Validations
        self._validate_domain()

        # Initialize adapters after loading domain
        self._initialize()

    def _traverse(self):
        # Standard Library Imports
        import importlib.util
        import os
        import pathlib

        root_dir = pathlib.PurePath(pathlib.Path(self.root_path).resolve()).parent
        path = pathlib.Path(root_dir)  # Resolve the domain file's directory
        system_folder_path = (
            path.parent
        )  # Get the directory of the domain file to traverse from

        logger.debug(f"Loading domain from {root_dir}...")

        # Identify subdirectories
        subdirectories = [
            name
            for name in os.listdir(root_dir)
            if os.path.isdir(os.path.join(root_dir, name))
            and name not in ["__pycache__"]
        ]

        # Identify subdirectories that have a toml file
        subdirectories_to_traverse = []
        files_to_check = ["domain.toml", ".domain.toml", "pyproject.toml"]
        for subdirectory in subdirectories:
            subdirectory_path = os.path.join(root_dir, subdirectory)
            if not any(
                file
                for file in files_to_check
                if os.path.isfile(os.path.join(subdirectory_path, file))
            ):
                subdirectories_to_traverse.append(subdirectory_path)

        # Traverse root directory
        for filename in os.listdir(root_dir):
            full_file_path = os.path.join(root_dir, filename)
            if (
                os.path.isfile(full_file_path)
                and os.path.splitext(filename)[1] == ".py"
                and full_file_path != self.root_path
            ):
                spec = importlib.util.spec_from_file_location(
                    filename, os.path.join(root_dir, filename)
                )
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)

                logger.debug(f"Loaded {filename}")

        # Traverse subdirectories recursively
        for directory in subdirectories_to_traverse:
            for root, _, files in os.walk(directory):
                if pathlib.PurePath(root).name not in ["__pycache__"]:
                    package_path = root[len(str(system_folder_path)) + 1 :]
                    module_name = package_path.replace(os.sep, ".")

                    for file in files:
                        file_base_name = os.path.basename(file)

                        # Ignore if the file is not a python file
                        if os.path.splitext(file_base_name)[1] != ".py":
                            continue

                        # Construct the module path to import from
                        if file_base_name != "__init__":
                            sub_module_name = os.path.splitext(file_base_name)[0]
                            file_module_name = module_name + "." + sub_module_name
                        else:
                            file_module_name = module_name
                        full_file_path = os.path.join(root, file)

                        try:
                            if (
                                full_file_path != self.root_path
                            ):  # Don't load the domain file itself again
                                spec = importlib.util.spec_from_file_location(
                                    file_module_name, full_file_path
                                )
                                module = importlib.util.module_from_spec(spec)
                                spec.loader.exec_module(module)

                                logger.debug(f"Loaded {file_module_name}")
                        except ModuleNotFoundError as exc:
                            logger.error(f"Error while loading a module: {exc}")

    def _initialize(self):
        """Initialize domain dependencies and adapters."""
        self.providers._initialize()
        self.caches._initialize()
        self.brokers._initialize()

    def make_config(self):
        """Used to construct the config; invoked by the Domain constructor."""
        defaults = dict(self.default_config)
        defaults["env"] = get_env()
        defaults["debug"] = get_debug_flag()
        return self.config_class(self.root_path, defaults)

    def load_config(self, load_toml=True):
        """Load configuration from dist or a .toml file."""
        if load_toml:
            return Config2.load_from_path(self.root_path, dict(self.default_config))
        else:
            return Config2.load_from_dict(dict(self.default_config))

    def domain_context(self, **kwargs):
        """Create an :class:`~protean.context.DomainContext`. Use as a ``with``
        block to push the context, which will make :data:`current_domain`
        point at this domain.

        ::

            with domain.domain_context():
                init_db()
        """
        return DomainContext(self, **kwargs)

    def teardown_domain_context(self, f):
        """Registers a function to be called when the domain context
        ends.

        Example::

            ctx = domain.domain_context()
            ctx.push()
            ...
            ctx.pop()

        When ``ctx.pop()`` is executed in the above example, the teardown
        functions are called just before the domain context moves from the
        stack of active contexts.  This becomes relevant if you are using
        such constructs in tests.

        When a teardown function was called because of an unhandled exception
        it will be passed an error object. If an :meth:`errorhandler` is
        registered, it will handle the exception and the teardown will not
        receive it.

        The return values of teardown functions are ignored.
        """
        self.teardown_domain_context_functions.append(f)
        return f

    def do_teardown_domain_context(self, exc=_sentinel):
        """Called right before the domain context is popped.

        This calls all functions decorated with
        :meth:`teardown_domain_context`.

        This is called by
        :meth:`DomainContext.pop() <protean.context.DomainContext.pop>`.
        """
        if exc is _sentinel:
            exc = sys.exc_info()[1]
        for func in reversed(self.teardown_domain_context_functions):
            func(exc)

    def __str__(self) -> str:
        return f"Domain: {self.name}"

    @property
    @lru_cache()
    def registry(self):
        return self._domain_registry

    def factory_for(self, domain_object_type):
        from protean.core.aggregate import aggregate_factory
        from protean.core.application_service import application_service_factory
        from protean.core.command import command_factory
        from protean.core.command_handler import command_handler_factory
        from protean.core.domain_service import domain_service_factory
        from protean.core.email import email_factory
        from protean.core.entity import entity_factory
        from protean.core.event import domain_event_factory
        from protean.core.event_handler import event_handler_factory
        from protean.core.event_sourced_aggregate import event_sourced_aggregate_factory
        from protean.core.model import model_factory
        from protean.core.repository import repository_factory
        from protean.core.serializer import serializer_factory
        from protean.core.subscriber import subscriber_factory
        from protean.core.value_object import value_object_factory
        from protean.core.view import view_factory

        factories = {
            DomainObjects.AGGREGATE.value: aggregate_factory,
            DomainObjects.APPLICATION_SERVICE.value: application_service_factory,
            DomainObjects.COMMAND.value: command_factory,
            DomainObjects.COMMAND_HANDLER.value: command_handler_factory,
            DomainObjects.EVENT.value: domain_event_factory,
            DomainObjects.EVENT_HANDLER.value: event_handler_factory,
            DomainObjects.EVENT_SOURCED_AGGREGATE.value: event_sourced_aggregate_factory,
            DomainObjects.DOMAIN_SERVICE.value: domain_service_factory,
            DomainObjects.EMAIL.value: email_factory,
            DomainObjects.ENTITY.value: entity_factory,
            DomainObjects.MODEL.value: model_factory,
            DomainObjects.REPOSITORY.value: repository_factory,
            DomainObjects.SUBSCRIBER.value: subscriber_factory,
            DomainObjects.SERIALIZER.value: serializer_factory,
            DomainObjects.VALUE_OBJECT.value: value_object_factory,
            DomainObjects.VIEW.value: view_factory,
        }

        if domain_object_type.value not in factories:
            raise IncorrectUsageError(
                {"_entity": [f"Unknown Element Type `{domain_object_type.value}`"]}
            )

        return factories[domain_object_type.value]

    def _register_element(self, element_type, element_cls, **kwargs):  # noqa: C901
        """Register class into the domain"""
        # Check if `element_cls` is already a subclass of the Element Type
        #   which would be the case in an explicit declaration like `class Account(BaseEntity):`
        #
        # We will need to construct a class derived from the right base class
        #   if the Element was specified through annotation, like so:
        #
        #  ```
        #       @Entity
        #       class Account:
        #  ```

        factory = self.factory_for(element_type)
        new_cls = factory(element_cls, **kwargs)

        if element_type == DomainObjects.MODEL:
            # Remember model association with aggregate/entity class, for easy fetching
            self._models[fqn(new_cls.meta_.entity_cls)] = new_cls

        # Register element with domain
        self._domain_registry.register_element(new_cls)

        # Resolve or record elements to be resolved

        # 1. Associations
        if has_fields(new_cls):
            for _, field_obj in declared_fields(new_cls).items():
                if isinstance(field_obj, (HasOne, HasMany, Reference)) and isinstance(
                    field_obj.to_cls, str
                ):
                    self._pending_class_resolutions[field_obj.to_cls].append(
                        ("Association", (field_obj, new_cls))
                    )

                if isinstance(field_obj, ValueObject) and isinstance(
                    field_obj.value_object_cls, str
                ):
                    self._pending_class_resolutions[field_obj.value_object_cls].append(
                        ("ValueObject", (field_obj, new_cls))
                    )

        # 2. Meta Linkages
        if element_type in [
            DomainObjects.ENTITY,
            DomainObjects.EVENT,
            DomainObjects.EVENT_HANDLER,
            DomainObjects.COMMAND,
            DomainObjects.COMMAND_HANDLER,
            DomainObjects.REPOSITORY,
            DomainObjects.EVENT_SOURCED_REPOSITORY,
        ]:
            if isinstance(new_cls.meta_.part_of, str):
                self._pending_class_resolutions[new_cls.meta_.part_of].append(
                    ("AggregateCls", (new_cls))
                )

        return new_cls

    def _resolve_references(self):
        """Resolve pending class references in association fields.

        Called by the domain context when domain is activated.
        """
        for name in list(self._pending_class_resolutions.keys()):
            for resolution_type, params in self._pending_class_resolutions[name]:
                if resolution_type == "Association":
                    field_obj, owner_cls = params
                    to_cls = self.fetch_element_cls_from_registry(
                        field_obj.to_cls,
                        (
                            DomainObjects.AGGREGATE,
                            DomainObjects.EVENT_SOURCED_AGGREGATE,
                            DomainObjects.ENTITY,
                        ),
                    )
                    field_obj._resolve_to_cls(self, to_cls, owner_cls)
                elif resolution_type == "ValueObject":
                    field_obj, owner_cls = params
                    to_cls = self.fetch_element_cls_from_registry(
                        field_obj.value_object_cls,
                        (DomainObjects.VALUE_OBJECT,),
                    )
                    field_obj._resolve_to_cls(self, to_cls, owner_cls)
                elif resolution_type == "AggregateCls":
                    cls = params
                    to_cls = self.fetch_element_cls_from_registry(
                        cls.meta_.part_of,
                        (
                            DomainObjects.AGGREGATE,
                            DomainObjects.EVENT_SOURCED_AGGREGATE,
                        ),
                    )
                    cls.meta_.part_of = to_cls

            # Remove from pending list now that the class has been resolved
            del self._pending_class_resolutions[name]

    # _cls should never be specified by keyword, so start it with an
    # underscore.  The presence of _cls is used to detect if this
    # decorator is being called with parameters or not.
    def _domain_element(
        self,
        element_type,
        _cls=None,
        **kwargs,
    ):
        """Returns the registered class after decoarating it and recording its presence in the domain"""

        def wrap(cls):
            return self._register_element(element_type, cls, **kwargs)

        # See if we're being called as @Entity or @Entity().
        if _cls is None:
            # We're called with parens.
            return wrap

        # We're called as @dataclass without parens.
        return wrap(_cls)

    def register_model(self, model_cls, **kwargs):
        """Register a model class"""
        return self._register_element(DomainObjects.MODEL, model_cls, **kwargs)

    def register(self, element_cls: Any, **kwargs: dict) -> Any:
        """Register an element with the domain.

        Returns the fully-formed domain element, subclassed and associated with
        all traits necessary for its element type.
        """

        # Reject unknown Domain Elements, identified by the absence of `element_type` class var
        if getattr(element_cls, "element_type", None) not in [
            element for element in DomainObjects
        ]:
            raise NotImplementedError

        return self._register_element(element_cls.element_type, element_cls, **kwargs)

    def delist(self, element_cls):
        """Delist a Domain Element.

        This method will result in a no-op if the entity class was not found
        in the registry for whatever reason.
        """
        if getattr(element_cls, "element_type", None) not in [
            element for element in DomainObjects
        ]:
            raise NotImplementedError

        self._domain_registry.dei_element(element_cls.element_type, element_cls)

    def fetch_element_cls_from_registry(
        self, element: Union[str, Any], element_types: Tuple[DomainObjects, ...]
    ) -> Any:
        """Util Method to fetch an Element's class from its name"""
        if isinstance(element, str):
            try:
                # Try fetching by class name
                return self._get_element_by_name(element_types, element).cls
            except ConfigurationError:
                try:
                    # Try fetching by fully qualified class name
                    return self._get_element_by_fully_qualified_name(
                        element_types, element
                    ).cls
                except ConfigurationError:
                    # Element has not been registered
                    # FIXME print a helpful debug message
                    raise
        else:
            # FIXME Check if entity is subclassed from BaseEntity
            return element

    def _get_element_by_name(self, element_types, element_name):
        """Fetch Domain record with the provided Element name"""
        try:
            elements = self._domain_registry._elements_by_name[element_name]

            # There is one element registered with the name and the correct type
            if len(elements) == 1 and elements[0].class_type in [
                o.value for o in element_types
            ]:
                return elements[0]
            else:
                # There are multiple elements registered with the name
                #   Loop to check if one of them has the correct type
                for element in elements:
                    if element.class_type in [o.value for o in element_types]:
                        return element
                else:
                    raise ConfigurationError(
                        {
                            "element": f"Element {element_name} not registered in domain {self.name}"
                        }
                    )
        except KeyError:
            raise ConfigurationError(
                {
                    "element": f"Element {element_name} not registered in domain {self.name}"
                }
            )

    def _get_element_by_fully_qualified_name(self, element_types, element_fq_name):
        """Fetch Domain record with the Fully Qualified Element name"""
        for element_type in element_types:
            if element_fq_name in self._domain_registry._elements[element_type.value]:
                return self._domain_registry._elements[element_type.value][
                    element_fq_name
                ]
        else:
            raise ConfigurationError(
                {
                    "element": f"Element {element_fq_name} not registered in domain {self.name}"
                }
            )

    def _get_element_by_class(self, element_types, element_cls):
        """Fetch Domain record with Element class details"""
        element_qualname = fqn(element_cls)
        return self._get_element_by_fully_qualified_name(
            element_types, element_qualname
        )

    def _replace_element_by_class(self, new_element_cls):
        aggregate_record = self._get_element_by_class(
            (DomainObjects.AGGREGATE, DomainObjects.ENTITY), new_element_cls
        )

        self._domain_registry._elements[aggregate_record.class_type][
            aggregate_record.qualname
        ].cls = new_element_cls

    def _validate_domain(self):
        """A method to validate the domain for correctness, called just before the domain is activated."""
        # Check `identity_function` is provided if `identity_strategy` is `function`
        if self.config["identity_strategy"] == self.IdentityStrategy.FUNCTION.value:
            if not self._identity_function:
                raise ConfigurationError(
                    {
                        "element": "Identity Strategy is set to `function`, but no Identity Function is provided"
                    }
                )

        # Check if all references are resolved
        if self._pending_class_resolutions:
            raise ConfigurationError(
                {
                    "element": f"Unresolved references in domain {self.name}",
                    "unresolved": self._pending_class_resolutions,
                }
            )

        # Confirm `HasOne` and `HasMany` fields are linked to entities and not aggregates
        for _, aggregate in self.registry.aggregates.items():
            for _, field_obj in declared_fields(aggregate.cls).items():
                if isinstance(field_obj, (HasOne, HasMany)):
                    if isinstance(field_obj.to_cls, str):
                        raise IncorrectUsageError(
                            {
                                "element": (
                                    f"Unresolved target `{field_obj.to_cls}` for field "
                                    f"`{aggregate.__name__}:{field_obj.name}`"
                                )
                            }
                        )
                    if field_obj.to_cls.element_type != DomainObjects.ENTITY:
                        raise IncorrectUsageError(
                            {
                                "element": (
                                    f"Field `{field_obj.field_name}` in `{aggregate.cls.__name__}` "
                                    "is not linked to an Entity class"
                                )
                            }
                        )

        # Check that no two event sourced aggregates have the same event class in their
        #   `_events_cls_map`.
        event_sourced_aggregates = self.registry._elements[
            DomainObjects.EVENT_SOURCED_AGGREGATE.value
        ]
        # Collect all event class names from `_events_cls_map` of all event sourced aggregates
        event_class_names = list()
        for event_sourced_aggregate in event_sourced_aggregates.values():
            event_class_names.extend(event_sourced_aggregate.cls._events_cls_map.keys())
        # Check for duplicates
        duplicate_event_class_names = set(
            [
                event_class_name
                for event_class_name in event_class_names
                if event_class_names.count(event_class_name) > 1
            ]
        )
        if len(duplicate_event_class_names) != 0:
            raise IncorrectUsageError(
                {
                    "_event": [
                        f"Events are associated with multiple event sourced aggregates: "
                        f"{', '.join(duplicate_event_class_names)}"
                    ]
                }
            )

    def _assign_aggregate_clusters(self):
        """Assign Aggregate Clusters to all relevant elements"""
        from protean.core.aggregate import BaseAggregate
        from protean.core.event_sourced_aggregate import BaseEventSourcedAggregate

        # Assign Aggregates and EventSourcedAggregates to their own cluster
        for element_type in [
            DomainObjects.AGGREGATE,
            DomainObjects.EVENT_SOURCED_AGGREGATE,
        ]:
            for _, element in self.registry._elements[element_type.value].items():
                element.cls.meta_.aggregate_cluster = element.cls

        # Derive root aggregate for other elements and assign as aggregate_cluster
        for element_type in [
            DomainObjects.ENTITY,
            DomainObjects.EVENT,
            DomainObjects.COMMAND,
        ]:
            for _, element in self.registry._elements[element_type.value].items():
                part_of = element.cls.meta_.part_of
                if part_of:
                    while not issubclass(
                        part_of, (BaseAggregate, BaseEventSourcedAggregate)
                    ):
                        part_of = part_of.meta_.part_of

                element.cls.meta_.aggregate_cluster = part_of

    ######################
    # Element Decorators #
    ######################

    def aggregate(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.AGGREGATE,
            _cls=_cls,
            **kwargs,
        )

    def application_service(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.APPLICATION_SERVICE,
            _cls=_cls,
            **kwargs,
        )

    def command(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.COMMAND,
            _cls=_cls,
            **kwargs,
        )

    def command_handler(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.COMMAND_HANDLER, _cls=_cls, **kwargs)

    def event(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.EVENT,
            _cls=_cls,
            **kwargs,
        )

    def event_handler(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.EVENT_HANDLER,
            _cls=_cls,
            **kwargs,
        )

    def event_sourced_aggregate(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.EVENT_SOURCED_AGGREGATE,
            _cls=_cls,
            **kwargs,
        )

    def domain_service(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.DOMAIN_SERVICE,
            _cls=_cls,
            **kwargs,
        )

    def entity(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.ENTITY, _cls=_cls, **kwargs)

    def email(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.EMAIL, _cls=_cls, **kwargs)

    def model(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.MODEL, _cls=_cls, **kwargs)

    def repository(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.REPOSITORY, _cls=_cls, **kwargs)

    def serializer(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.SERIALIZER, _cls=_cls, **kwargs)

    def subscriber(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.SUBSCRIBER,
            _cls=_cls,
            **kwargs,
        )

    def value_object(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.VALUE_OBJECT,
            _cls=_cls,
            **kwargs,
        )

    def view(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.VIEW,
            _cls=_cls,
            **kwargs,
        )

    ########################
    # Broker Functionality #
    ########################

    def publish(self, events: Union[BaseEvent, List[BaseEvent]]) -> None:
        """Publish Events to all configured brokers.
        Args:
            events (BaseEvent): The Event object containing data to be pushed
        """
        if not isinstance(events, list):
            events = [events]

        for event in events:
            # Persist event in Message Store
            self.event_store.store.append(event)

            self.brokers.publish(event)

            if self.config["event_processing"] == EventProcessing.SYNC.value:
                # Consume events right-away
                handler_classes = self.handlers_for(event)
                for handler_cls in handler_classes:
                    handler_methods = (
                        handler_cls._handlers[fqn(event.__class__)]
                        or handler_cls._handlers["$any"]
                    )

                    for handler_method in handler_methods:
                        handler_method(handler_cls(), event)

    #####################
    # Handling Commands #
    #####################
    def process(self, command: BaseCommand, asynchronous: bool = True) -> Optional[Any]:
        """Process command and return results based on specified preference.

        By default, Protean does not return values after processing commands. This behavior
        can be overridden either by setting command_processing in config to "sync" or by specifying
        ``asynchronous=False`` when calling the domain's ``handle`` method.

        Args:
            command (BaseCommand): Command to process
            asynchronous (Boolean, optional): Specifies if the command should be processed asynchronously.
                Defaults to True.

        Returns:
            Optional[Any]: Returns either the command handler's return value or nothing, based on preference.
        """
        position = self.event_store.store.append(command)

        if (
            not asynchronous
            or self.config["command_processing"] == CommandProcessing.SYNC.value
        ):
            handler_class = self.command_handler_for(command)
            if handler_class:
                handler_method = next(
                    iter(handler_class._handlers[fqn(command.__class__)])
                )
                handler_method(handler_class(), command)

        return position

    def command_handler_for(self, command: BaseCommand) -> Optional[BaseCommandHandler]:
        """Return Command Handler for a specific command.

        Args:
            command (BaseCommand): Command to process

        Returns:
            Optional[BaseCommandHandler]: Command Handler registered to process the command
        """
        return self.event_store.command_handler_for(command)

    ###################
    # Handling Events #
    ###################
    def raise_(self, event: BaseEvent) -> None:
        """Raise Domain Event in the stream specified in its options, or using the associated
        aggregate's stream name.

        Args:
            event (BaseEvent): Event to record

        Returns:
            None: Returns nothing.
        """
        self.event_store.store.append(event)

    def handlers_for(self, event: BaseEvent) -> List[BaseEventHandler]:
        """Return Event Handlers listening to a specific event

        Args:
            event (BaseEvent): Event to be consumed

        Returns:
            List[BaseEventHandler]: Event Handlers that have registered to consume the event
        """
        return self.event_store.handlers_for(event)

    ############################
    # Repository Functionality #
    ############################

    # FIXME Optimize calls to this method with cache, but also with support for Multitenancy
    def repository_for(self, part_of) -> BaseRepository:
        if isinstance(part_of, str):
            raise IncorrectUsageError(
                {
                    "element": [
                        f"Element {part_of} is not registered in domain {self.name}"
                    ]
                }
            )

        if part_of.element_type == DomainObjects.EVENT_SOURCED_AGGREGATE:
            return self.event_store.repository_for(part_of)
        else:
            return self.providers.repository_for(part_of)

    #######################
    # Cache Functionality #
    #######################

    def cache_for(self, view_cls):
        return self.caches.cache_for(view_cls)

    #######################
    # Email Functionality #
    #######################

    def get_email_provider(self, provider_name):
        return self.email_providers.get_email_provider(provider_name)

    def send_email(self, email):
        return self.email_providers.send_email(email)

    def make_shell_context(self):
        """Return a dictionary of context variables for a shell session."""
        values = {"domain": self}

        # For each domain element type in Domain Objects,
        #   Cycle through all values in self.registry._elements[element_type]
        #   and add each class to the shell context by the key
        for element_type in DomainObjects:
            values.update(
                {
                    v.name: v.cls
                    for _, v in self._domain_registry._elements[
                        element_type.value
                    ].items()
                }
            )

        return values



__version__ = "0.12.0"

from .core.aggregate import BaseAggregate, atomic_change
from .core.application_service import BaseApplicationService
from .core.command import BaseCommand
from .core.command_handler import BaseCommandHandler
from .core.domain_service import BaseDomainService
from .core.email import BaseEmail
from .core.entity import BaseEntity, invariant
from .core.event import BaseEvent
from .core.event_handler import BaseEventHandler
from .core.event_sourced_aggregate import BaseEventSourcedAggregate, apply
from .core.model import BaseModel
from .core.queryset import Q, QuerySet
from .core.repository import BaseRepository
from .core.serializer import BaseSerializer
from .core.subscriber import BaseSubscriber
from .core.unit_of_work import UnitOfWork
from .core.value_object import BaseValueObject
from .core.view import BaseView
from .domain import Domain
from .server import Engine
from .utils import get_version
from .utils.mixins import handle

__all__ = [
    "BaseAggregate",
    "BaseApplicationService",
    "BaseCommand",
    "BaseCommandHandler",
    "BaseDomainService",
    "BaseEmail",
    "BaseEntity",
    "BaseEvent",
    "BaseEventHandler",
    "BaseEventSourcedAggregate",
    "BaseModel",
    "BaseRepository",
    "BaseSerializer",
    "BaseSubscriber",
    "BaseValueObject",
    "BaseView",
    "Domain",
    "Engine",
    "Q",
    "QuerySet",
    "UnitOfWork",
    "apply",
    "get_version",
    "handle",
    "invariant",
    "atomic_change",
]



from __future__ import annotations

import functools
import logging
from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, Union
from uuid import uuid4

from protean import fields
from protean.container import BaseContainer, OptionsMixin
from protean.core.command import BaseCommand
from protean.core.event import BaseEvent
from protean.core.event_sourced_aggregate import BaseEventSourcedAggregate
from protean.core.unit_of_work import UnitOfWork
from protean.core.value_object import BaseValueObject
from protean.exceptions import ConfigurationError
from protean.globals import current_domain, g
from protean.reflection import has_id_field, id_field
from protean.utils import fully_qualified_name

logger = logging.getLogger(__name__)


class MessageType(Enum):
    EVENT = "EVENT"
    COMMAND = "COMMAND"
    READ_POSITION = "READ_POSITION"


class MessageMetadata(BaseValueObject):
    # Marks message as a `COMMAND` or an `EVENT`
    kind = fields.String(required=True, max_length=15, choices=MessageType)

    # Name of service that owns the contract of the message
    owner = fields.String(max_length=50)

    # Allows for parsing of different versions, in case of
    #   breaking changes.
    schema_version = fields.Integer()

    # `origin_stream_name` helps keep track of the origin of a message.
    #   A command created by an event is automatically associated with the original stream.
    #   Events raised subsequently by the commands also carry forward the original stream name.
    origin_stream_name = fields.String()

    # FIXME Provide mechanism to add custom metadata fields/structure
    #   Can come handy in case of multi-tenancy, etc.


class MessageRecord(BaseContainer):
    """
    Base Container holding all fields of a message.
    """

    # Primary key. The ordinal position of the message in the entire message store.
    # Global position may have gaps.
    global_position = fields.Auto(increment=True, identifier=True)

    # The ordinal position of the message in its stream.
    # Position is gapless.
    position = fields.Integer()

    # Message creation time
    time = fields.DateTime()

    # Unique ID of the message
    id = fields.Auto()

    # Name of stream to which the message is written
    stream_name = fields.String(max_length=255)

    # The type of the message
    type = fields.String()

    # JSON representation of the message body
    data = fields.Dict()

    # JSON representation of the message metadata
    metadata = fields.ValueObject(MessageMetadata)


class Message(MessageRecord, OptionsMixin):  # FIXME Remove OptionsMixin
    """Generic message class
    It provides concrete implementations for:
    - ID generation
    - Payload construction
    - Serialization and De-serialization
    """

    # Version that the stream is expected to be when the message is written
    expected_version = fields.Integer()

    @classmethod
    def derived_metadata(cls, new_message_type: str) -> Dict:
        additional_metadata = {}

        if hasattr(g, "message_in_context"):
            if (
                new_message_type == "COMMAND"
                and g.message_in_context.metadata.kind == "EVENT"
            ):
                additional_metadata["origin_stream_name"] = (
                    g.message_in_context.stream_name
                )

            if (
                new_message_type == "EVENT"
                and g.message_in_context.metadata.kind == "COMMAND"
                and g.message_in_context.metadata.origin_stream_name is not None
            ):
                additional_metadata["origin_stream_name"] = (
                    g.message_in_context.metadata.origin_stream_name
                )
        return additional_metadata

    @classmethod
    def from_dict(cls, message: Dict) -> Message:
        return Message(
            stream_name=message["stream_name"],
            type=message["type"],
            data=message["data"],
            metadata=MessageMetadata(**message["metadata"]),
            position=message["position"],
            global_position=message["global_position"],
            time=message["time"],
            id=message["id"],
        )

    @classmethod
    def to_aggregate_event_message(
        cls, aggregate: BaseEventSourcedAggregate, event: BaseEvent
    ) -> Message:
        identifier = getattr(aggregate, id_field(aggregate).field_name)

        # Take the Aggregate's stream_name
        aggregate_stream_name = None
        if event.meta_.aggregate_cluster:
            aggregate_stream_name = event.meta_.aggregate_cluster.meta_.stream_name

        # Use explicit stream name if provided, or fallback on Aggregate's stream name
        stream_name = event.meta_.stream_name or aggregate_stream_name

        return cls(
            stream_name=f"{stream_name}-{identifier}",
            type=fully_qualified_name(event.__class__),
            data=event.to_dict(),
            metadata=MessageMetadata(
                kind=MessageType.EVENT.value,
                owner=current_domain.name,
                **cls.derived_metadata(MessageType.EVENT.value),
                # schema_version=event.meta_.version,  # FIXME Maintain version for event
            ),
            expected_version=aggregate._version,  # FIXME Maintain version for Aggregates
        )

    def to_object(self) -> Union[BaseEvent, BaseCommand]:
        if self.metadata.kind == MessageType.EVENT.value:
            element_record = current_domain.registry.events[self.type]
        elif self.metadata.kind == MessageType.COMMAND.value:
            element_record = current_domain.registry.commands[self.type]
        else:
            raise NotImplementedError  # FIXME Handle unknown messages better

        if not element_record:
            raise ConfigurationError(
                f"Element {self.type.split('.')[-1]} is not registered with the domain"
            )

        return element_record.cls(**self.data)

    @classmethod
    def to_message(cls, message_object: Union[BaseEvent, BaseCommand]) -> Message:
        if has_id_field(message_object):
            identifier = getattr(message_object, id_field(message_object).field_name)
        else:
            identifier = str(uuid4())

        # Take the Aggregate's stream_name
        aggregate_stream_name = None
        if message_object.meta_.aggregate_cluster:
            aggregate_stream_name = (
                message_object.meta_.aggregate_cluster.meta_.stream_name
            )

        # Use explicit stream name if provided, or fallback on Aggregate's stream name
        stream_name = message_object.meta_.stream_name or aggregate_stream_name

        if isinstance(message_object, BaseEvent):
            stream_name = f"{stream_name}-{identifier}"
            kind = MessageType.EVENT.value
        elif isinstance(message_object, BaseCommand):
            stream_name = f"{stream_name}:command-{identifier}"
            kind = MessageType.COMMAND.value
        else:
            raise NotImplementedError  # FIXME Handle unknown messages better

        return cls(
            stream_name=stream_name,
            type=fully_qualified_name(message_object.__class__),
            data=message_object.to_dict(),
            metadata=MessageMetadata(
                kind=kind,
                owner=current_domain.name,
                **cls.derived_metadata(kind),
            ),
            # schema_version=command.meta_.version,  # FIXME Maintain version
        )


class handle:
    """Class decorator to mark handler methods in EventHandler and CommandHandler classes."""

    def __init__(self, target_cls: Union[BaseEvent, BaseCommand]) -> None:
        self._target_cls = target_cls

    def __call__(self, fn: Callable) -> Callable:
        """Marks the method with a special `_target_cls` attribute to be able to
        construct a map of handlers later.

        Args:
            fn (Callable): Handler method

        Returns:
            Callable: Handler method with `_target_cls` attribute
        """

        @functools.wraps(fn)
        def wrapper(instance, target_obj):
            # Wrap function call within a UoW
            with UnitOfWork():
                fn(instance, target_obj)

        setattr(wrapper, "_target_cls", self._target_cls)
        return wrapper


class HandlerMixin:
    """Mixin to add common handler behavior to Event Handlers and Command Handlers"""

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        # Associate a `_handlers` map with subclasses.
        #   It needs to be initialized here because if it
        #   were initialized in __init__, the same collection object
        #   would be made available across all subclasses,
        #   defeating its purpose.
        setattr(subclass, "_handlers", defaultdict(set))

    @classmethod
    def _handle(cls, message: Message) -> None:
        # Use Event-specific handlers if available, or fallback on `$any` if defined
        handlers = cls._handlers[message.type] or cls._handlers["$any"]

        for handler_method in handlers:
            handler_method(cls(), message.to_object())

    @classmethod
    def handle_error(cls, exc: Exception, message: Message) -> None:
        """Default error handler for messages. Can be overridden in subclasses.

        By default, this method logs the error and raises it.
        """



"""Entity Functionality and Classes"""

import copy
import functools
import inspect
import logging
from collections import defaultdict
from functools import partial

from protean.container import BaseContainer, IdentityMixin, OptionsMixin
from protean.exceptions import IncorrectUsageError, NotSupportedError, ValidationError
from protean.fields import Auto, HasMany, Reference, ValueObject
from protean.fields.association import Association
from protean.reflection import _FIELDS, attributes, declared_fields, fields, id_field
from protean.utils import (
    DomainObjects,
    derive_element_class,
    generate_identity,
    inflection,
)

logger = logging.getLogger(__name__)


class _FieldsCacheDescriptor:
    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        res = instance.fields_cache = {}
        return res


class _EntityState:
    """Store entity instance state."""

    def __init__(self):
        self._new = True
        self._changed = False
        self._destroyed = False

    @property
    def is_new(self):
        return self._new

    @property
    def is_persisted(self):
        return not self._new

    @property
    def is_changed(self):
        return self._changed

    @property
    def is_destroyed(self):
        return self._destroyed

    def mark_new(self):
        self._new = True

    def mark_saved(self):
        self._new = False
        self._changed = False

    mark_retrieved = (
        mark_saved  # Alias as placeholder so that future change wont affect interface
    )

    def mark_changed(self):
        if not (self._new or self._destroyed):
            self._changed = True

    def mark_destroyed(self):
        self._destroyed = True
        self._changed = False

    fields_cache = _FieldsCacheDescriptor()


class BaseEntity(OptionsMixin, IdentityMixin, BaseContainer):
    """The Base class for Protean-Compliant Domain Entities.

    Provides helper methods to custom define entity attributes, and query attribute names
    during runtime.

    You can define an Entity with the help of `entity` annotation::

        @domain.entity
        class User:
            id = field.Integer(identifier=True)
            first_name = field.String(required=True, max_length=50)
            last_name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)

    (or)

    Or, you can directly subclass from `BaseEntity`::

        class User(BaseEntity):
            id = field.Integer(identifier=True)
            first_name = field.String(required=True, max_length=50)
            last_name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)

        domain.register_element(User)

    During persistence, the model associated with this entity is retrieved dynamically from the repository factory,
    initialized with a live connection to the datastore.
    """

    element_type = DomainObjects.ENTITY

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        # Record invariant methods
        setattr(subclass, "_invariants", defaultdict(dict))

    @classmethod
    def _default_options(cls):
        return [
            ("auto_add_id_field", True),
            ("provider", "default"),
            ("model", None),
            ("part_of", None),
            ("aggregate_cluster", None),
            ("schema_name", inflection.underscore(cls.__name__)),
        ]

    @classmethod
    def _extract_options(cls, **opts):
        """A stand-in method for setting customized options on the Domain Element

        Empty by default. To be overridden in each Element that expects or needs
        specific options.
        """
        for key, default in cls._default_options():
            value = (
                opts.pop(key, None)
                or (hasattr(cls.meta_, key) and getattr(cls.meta_, key))
                or default
            )
            setattr(cls.meta_, key, value)

    def __init__(self, *template, **kwargs):  # noqa: C901
        """
        Initialise the entity object.

        During initialization, set value on fields if validation passes.

        This initialization technique supports keyword arguments as well as dictionaries. The objects initialized
        in the following example have the same structure::

            user1 = User({'first_name': 'John', 'last_name': 'Doe'})

            user2 = User(first_name='John', last_name='Doe')

        You can also specify a template for initial data and override specific attributes::

            base_user = User({'age': 15})

            user = User(base_user.to_dict(), first_name='John', last_name='Doe')
        """

        self._initialized = False

        if self.meta_.abstract is True:
            raise NotSupportedError(
                f"{self.__class__.__name__} class has been marked abstract"
                f" and cannot be instantiated"
            )

        self.errors = defaultdict(list)

        # Set up the storage for instance state
        self.state_ = _EntityState()

        # Placeholder for HasMany change tracking
        self._temp_cache = defaultdict(lambda: defaultdict(dict))

        # Attributes to preserve heirarchy of element instances
        self._owner = None
        self._root = None

        # To control invariant checks
        self._disable_invariant_checks = False

        # Placeholder for temporary storage of raised events
        self._events = []

        # Collect Reference field attribute names to prevent accidental overwriting
        # of shadow fields.
        reference_attributes = {
            field_obj.attribute_name: field_obj.field_name
            for field_obj in declared_fields(self).values()
            if isinstance(field_obj, Reference)
        }

        # Track fields that have been loaded
        loaded_fields = []

        # Pick identifier if provided in template or kwargs
        #   Generate if not provided
        #
        # Find identity field name
        id_field_obj = id_field(self)
        id_field_name = id_field_obj.field_name

        # Look for id field in kwargs and load value if present
        if kwargs and id_field_name in kwargs:
            setattr(self, id_field_name, kwargs.pop(id_field_name))
            loaded_fields.append(id_field_name)
        elif template:
            # Look for id field in template dictionary and load value if present
            for dictionary in template:
                if id_field_name in dictionary:
                    setattr(self, id_field_name, dictionary[id_field_name])
                    loaded_fields.append(id_field_name)
                    break
        else:
            # A value was not supplied, so generate one
            if type(id_field_obj) is Auto and not id_field_obj.increment:
                setattr(
                    self,
                    id_field_name,
                    generate_identity(
                        id_field_obj.identity_strategy,
                        id_field_obj.identity_function,
                        id_field_obj.identity_type,
                    ),
                )
                loaded_fields.append(id_field_name)

        # Load the attributes based on the template
        for dictionary in template:
            if not isinstance(dictionary, dict):
                raise AssertionError(
                    f'Positional argument "{dictionary}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in dictionary.items():
                if field_name not in kwargs and field_name not in loaded_fields:
                    kwargs[field_name] = val

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            if field_name not in loaded_fields:
                try:
                    setattr(self, field_name, val)
                except ValidationError as err:
                    for field_name in err.messages:
                        self.errors[field_name].extend(err.messages[field_name])
                finally:
                    loaded_fields.append(field_name)

                    # Also note reference field name if its attribute was loaded
                    if field_name in reference_attributes:
                        loaded_fields.append(reference_attributes[field_name])

        # Load Value Objects from associated fields
        #   This block will dynamically construct value objects from field values
        #   and associated the vo with the entity
        # If the value object was already provided, it will not be overridden.
        for field_name, field_obj in declared_fields(self).items():
            if isinstance(field_obj, (ValueObject)) and not getattr(self, field_name):
                attrs = [
                    (embedded_field.field_name, embedded_field.attribute_name)
                    for embedded_field in field_obj.embedded_fields.values()
                ]
                kwargs_values = {name: kwargs.get(attr) for name, attr in attrs}

                # Check if any of the values in `values` are not None
                #   If all values are None, it means that the value object is not being set
                #   and we should set it to None
                #
                #   If any of the values are not None, we should set the value object and its attributes
                #   to the values provided and let it trigger validations
                if any(kwargs_values.values()):
                    try:
                        value_object = field_obj.value_object_cls(**kwargs_values)

                        # Set VO value only if the value object is not None/Empty
                        if value_object:
                            setattr(self, field_name, value_object)
                            loaded_fields.append(field_name)
                    except ValidationError as err:
                        for sub_field_name in err.messages:
                            self.errors[
                                "{}_{}".format(field_name, sub_field_name)
                            ].extend(err.messages[sub_field_name])

        # Load other identities
        for field_name, field_obj in declared_fields(self).items():
            if (
                field_name not in loaded_fields
                and type(field_obj) is Auto
                and not field_obj.increment
            ):
                if not getattr(self, field_obj.field_name, None):
                    setattr(
                        self,
                        field_obj.field_name,
                        generate_identity(
                            field_obj.identity_strategy,
                            field_obj.identity_function,
                            field_obj.identity_type,
                        ),
                    )
                loaded_fields.append(field_obj.field_name)

        # Load Associations
        for field_name, field_obj in declared_fields(self).items():
            if isinstance(field_obj, Association):
                getattr(self, field_name)  # This refreshes the values in associations

                # Set up add and remove methods. These are pseudo methods: `add_*`,
                #   `remove_*` and `filter_*` that point to the HasMany field's `add`,
                #   `remove` and `filter` methods. They are wrapped to ensure we pass
                #   the object that holds the values and temp_cache.
                if isinstance(field_obj, HasMany):
                    setattr(self, f"add_{field_name}", partial(field_obj.add, self))
                    setattr(
                        self, f"remove_{field_name}", partial(field_obj.remove, self)
                    )
                    setattr(
                        self, f"get_one_from_{field_name}", partial(field_obj.get, self)
                    )
                    setattr(
                        self, f"filter_{field_name}", partial(field_obj.filter, self)
                    )

        self.defaults()

        # Now load the remaining fields with a None value, which will fail
        # for required fields
        for field_name, field_obj in fields(self).items():
            if field_name not in loaded_fields and (
                not hasattr(self, field_name) or getattr(self, field_name) is None
            ):
                if not isinstance(field_obj, Association):
                    try:
                        setattr(self, field_name, None)

                        # If field is a VO, set underlying attributes to None as well
                        if isinstance(field_obj, ValueObject):
                            for embedded_field in field_obj.embedded_fields.values():
                                setattr(self, embedded_field.attribute_name, None)
                    except ValidationError as err:
                        for field_name in err.messages:
                            self.errors[field_name].extend(err.messages[field_name])

        for field_name, field_obj in attributes(self).items():
            if field_name not in loaded_fields and not hasattr(self, field_name):
                setattr(self, field_name, None)

        self._initialized = True

        # `_postcheck()` will return a `defaultdict(list)` if errors are to be raised
        custom_errors = self._postcheck(return_errors=True) or {}
        for field in custom_errors:
            self.errors[field].extend(custom_errors[field])

        # Raise any errors found during load
        if self.errors:
            logger.error(f"Error during initialization: {dict(self.errors)}")
            raise ValidationError(self.errors)

    def defaults(self):
        """Placeholder method for defaults.
        To be overridden in concrete Containers, when an attribute's default depends on other attribute values.
        """

    def _run_invariants(self, stage, return_errors=False):
        """Run invariants for a given stage."""
        if self._initialized and not self._disable_invariant_checks:
            errors = defaultdict(list)

            for invariant_method in self._invariants[stage].values():
                try:
                    invariant_method(self)
                except ValidationError as err:
                    for field_name in err.messages:
                        errors[field_name].extend(err.messages[field_name])

            # Run through all associations and trigger their invariants
            for field_name, field_obj in declared_fields(self).items():
                if isinstance(field_obj, (Association, ValueObject)):
                    value = getattr(self, field_name)
                    if value is not None:
                        items = value if isinstance(value, list) else [value]
                        for item in items:
                            # Pre-checks don't apply to ValueObjects, because VOs are immutable
                            #   and therefore cannot be changed once initialized.
                            if stage == "pre" and not isinstance(
                                field_obj, ValueObject
                            ):
                                item_errors = item._precheck(return_errors=True)
                            else:
                                item_errors = item._postcheck(return_errors=True)
                            if item_errors:
                                for sub_field_name, error_list in item_errors.items():
                                    errors[sub_field_name].extend(error_list)

            if return_errors:
                return errors

            if errors:
                raise ValidationError(errors)

    def _precheck(self, return_errors=False):
        """Invariant checks performed before entity changes"""
        return self._run_invariants("pre", return_errors=return_errors)

    def _postcheck(self, return_errors=False):
        """Invariant checks performed after initialization and attribute changes"""
        return self._run_invariants("post", return_errors=return_errors)

    def raise_(self, event) -> None:
        """Raise an event in the aggregate cluster.

        The event is always registered on the aggregate, irrespective of where
        it is raised in the entity cluster."""
        self._root._events.append(event)

    def __eq__(self, other):
        """Equivalence check to be based only on Identity"""

        # FIXME Enhanced Equality Checks
        #   * Ensure IDs have values and both of them are not null
        #   * Ensure that the ID is of the right type
        #   * Ensure that Objects belong to the same `type`
        #   * Check Reference equality

        # FIXME Check if `==` and `in` operator work with __eq__

        if type(other) is type(self):
            self_id = getattr(self, id_field(self).field_name)
            other_id = getattr(other, id_field(other).field_name)

            return self_id == other_id

        return False

    def __hash__(self):
        """Overrides the default implementation and bases hashing on identity"""

        # FIXME Add Object Class Type to hash
        return hash(getattr(self, id_field(self).field_name))

    def _update_data(self, *data_dict, **kwargs):
        """
        A private method to process and update entity values correctly.

        :param data: A dictionary of values to be updated for the entity
        :param kwargs: keyword arguments with key-value pairs to be updated
        """

        # Load each of the fields given in the data dictionary
        self.errors = {}

        for data in data_dict:
            if not isinstance(data, dict):
                raise AssertionError(
                    f'Positional argument "{data}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in data.items():
                setattr(self, field_name, val)

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            setattr(self, field_name, val)

        # Raise any errors found during update
        if self.errors:
            logger.error(f"Errors on Update: {dict(self.errors)}")
            raise ValidationError(self.errors)

    def to_dict(self):
        """Return entity data as a dictionary"""
        # FIXME Memoize this function
        field_values = {}

        for field_name, field_obj in declared_fields(self).items():
            if (
                not isinstance(field_obj, (ValueObject, Reference))
                and getattr(self, field_name, None) is not None
            ):
                field_values[field_name] = field_obj.as_dict(
                    getattr(self, field_name, None)
                )
            elif isinstance(field_obj, ValueObject):
                value = field_obj.as_dict(getattr(self, field_name, None))
                if value:
                    field_values[field_name] = field_obj.as_dict(
                        getattr(self, field_name, None)
                    )

        return field_values

    def __repr__(self):
        """Friendly repr for Entity"""
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        identifier = getattr(self, id_field(self).field_name)
        return "%s object (%s)" % (
            self.__class__.__name__,
            "{}: {}".format(id_field(self).field_name, identifier),
        )

    def clone(self):
        """Deepclone the entity, but reset state"""
        clone_copy = copy.deepcopy(self)
        clone_copy.state_ = _EntityState()

        return clone_copy

    def _set_root_and_owner(self, root, owner):
        """Set the root and owner entities on all child entities

        This is a recursive process set in motion by the aggregate's `__init__` method.
        """
        self._root = root
        self._owner = owner

        # Set `_root` on all child entities
        for field_name, field_obj in declared_fields(self).items():
            # We care only about enclosed fields (associations)
            if isinstance(field_obj, Association):
                # Get current assigned value
                value = getattr(self, field_name)
                if value is not None:
                    # Link child entities to own root
                    items = value if isinstance(value, list) else [value]
                    for item in items:
                        if not item._root:
                            item._set_root_and_owner(self._root, self)


def entity_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseEntity, **kwargs)

    if not element_cls.meta_.part_of:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Entity `{element_cls.__name__}` needs to be associated with an Aggregate"
                ]
            }
        )

    # Set up reference fields
    if not element_cls.meta_.abstract:
        reference_field = None
        for field_obj in declared_fields(element_cls).values():
            if isinstance(field_obj, Reference):
                # An explicit `Reference` field is already present
                reference_field = field_obj
                break

        if reference_field is None:
            # If no explicit Reference field is present, create one
            reference_field = Reference(element_cls.meta_.part_of)

            # If part_of is a string, set field name to inflection.underscore(part_of)
            #   Else, if it is a class, extract class name and set field name to inflection.underscore(class_name)
            if isinstance(element_cls.meta_.part_of, str):
                field_name = inflection.underscore(element_cls.meta_.part_of)
            else:
                field_name = inflection.underscore(element_cls.meta_.part_of.__name__)

            setattr(element_cls, field_name, reference_field)

            # Set the name of the field on itself
            reference_field.__set_name__(element_cls, field_name)

            # FIXME Centralize this logic to add fields dynamically to _FIELDS
            field_objects = getattr(element_cls, _FIELDS)
            field_objects[field_name] = reference_field
            setattr(element_cls, _FIELDS, field_objects)

        # Set up shadow fields for Reference fields
        for _, field in fields(element_cls).items():
            if isinstance(field, Reference):
                shadow_field_name, shadow_field = field.get_shadow_field()
                shadow_field.__set_name__(element_cls, shadow_field_name)

    # Iterate through methods marked as `@invariant` and record them for later use
    methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
    for method_name, method in methods:
        if not (
            method_name.startswith("__") and method_name.endswith("__")
        ) and hasattr(method, "_invariant"):
            element_cls._invariants[method._invariant][method_name] = method

    return element_cls


class invariant:
    @staticmethod
    def pre(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(wrapper, "_invariant", "pre")
        return wrapper

    @staticmethod
    def post(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(wrapper, "_invariant", "post")
        return wrapper



import subprocess

import typer

app = typer.Typer(no_args_is_help=True)


"""
If we want to create a CLI app with one single command but
still want it to be a command/subcommand, we need to add a callback (see below).

This can be removed when we have more than one command/subcommand.

https://typer.tiangolo.com/tutorial/commands/one-or-multiple/#one-command-and-one-callback
"""


@app.callback()
def callback():
    pass


@app.command()
def preview():
    """Run a live preview server"""
    try:
        subprocess.call(["mkdocs", "serve"])
    except KeyboardInterrupt:
        pass



from abc import abstractmethod

from protean.container import Element, OptionsMixin
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.utils import DomainObjects, derive_element_class


class BaseModel(Element, OptionsMixin):
    """This is a Model representing a data schema in the persistence store. A concrete implementation of this
    model has to be provided by each persistence store plugin.
    """

    element_type = DomainObjects.MODEL

    def __new__(cls, *args, **kwargs):
        if cls is BaseModel:
            raise NotSupportedError("BaseModel cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    def _default_options(cls):
        return [
            ("entity_cls", None),
            ("schema_name", None),
            ("database", None),
        ]

    @classmethod
    @abstractmethod
    def from_entity(cls, entity):
        """Initialize Model object from Entity object"""

    @classmethod
    @abstractmethod
    def to_entity(cls, *args, **kwargs):
        """Convert Model Object to Entity Object"""


def model_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseModel, **kwargs)

    if not element_cls.meta_.entity_cls:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Model `{element_cls.__name__}` should be associated with an Entity or Aggregate"
                ]
            }
        )

    return element_cls



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mprotean` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``protean.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``protean.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import logging
import subprocess
from enum import Enum
from typing import Optional

import typer
from rich import print
from typing_extensions import Annotated

from protean.cli.docs import app as docs_app
from protean.cli.generate import app as generate_app
from protean.cli.new import new
from protean.cli.shell import shell
from protean.exceptions import NoDomainException
from protean.utils.domain_discovery import derive_domain

logger = logging.getLogger(__name__)

# Create the Typer app
#   `no_args_is_help=True` will show the help message when no arguments are passed
app = typer.Typer(no_args_is_help=True)

app.command()(new)
app.command()(shell)
app.add_typer(generate_app, name="generate")
app.add_typer(docs_app, name="docs")


class Category(str, Enum):
    CORE = "CORE"
    EVENTSTORE = "EVENTSTORE"
    DATABASE = "DATABASE"
    FULL = "FULL"


def version_callback(value: bool):
    if value:
        from protean import __version__

        typer.echo(f"Protean {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    ctx: typer.Context,
    version: Annotated[
        bool, typer.Option(help="Show version information", callback=version_callback)
    ] = False,
):
    """
    Protean CLI
    """


@app.command()
def test(
    category: Annotated[
        Category, typer.Option("-c", "--category", case_sensitive=False)
    ] = Category.CORE,
):
    commands = ["pytest", "--cache-clear", "--ignore=tests/support/"]

    match category.value:
        case "EVENTSTORE":
            # Run tests for EventStore adapters
            # FIXME: Add support for auto-fetching supported event stores
            for store in ["MEMORY", "MESSAGE_DB"]:
                print(f"Running tests for EVENTSTORE: {store}...")
                subprocess.call(commands + ["-m", "eventstore", f"--store={store}"])
        case "DATABASE":
            # Run tests for database adapters
            # FIXME: Add support for auto-fetching supported databases
            for db in ["POSTGRESQL", "SQLITE"]:
                print(f"Running tests for DATABASE: {db}...")
                subprocess.call(commands + ["-m", "database", f"--db={db}"])
        case "FULL":
            # Run full suite of tests with coverage
            # FIXME: Add support for auto-fetching supported adapters
            subprocess.call(
                commands
                + [
                    "--slow",
                    "--sqlite",
                    "--postgresql",
                    "--elasticsearch",
                    "--redis",
                    "--message_db",
                    "--cov=protean",
                    "--cov-config",
                    ".coveragerc",
                    "tests",
                ]
            )

            # Test against each supported database
            for db in ["POSTGRESQL", "SQLITE"]:
                print(f"Running tests for DB: {db}...")

                subprocess.call(commands + ["-m", "database", f"--db={db}"])

            for store in ["MESSAGE_DB"]:
                print(f"Running tests for EVENTSTORE: {store}...")
                subprocess.call(commands + ["-m", "eventstore", f"--store={store}"])
        case _:
            print("Running core tests...")
            subprocess.call(commands)


@app.command()
def server(
    domain: Annotated[str, typer.Option()] = ".",
    test_mode: Annotated[Optional[bool], typer.Option()] = False,
    debug: Annotated[Optional[bool], typer.Option()] = False,
):
    """Run Async Background Server"""
    # FIXME Accept MAX_WORKERS as command-line input as well
    try:
        domain = derive_domain(domain)
    except NoDomainException as exc:
        logger.error(f"Error loading Protean domain: {exc.messages}")
        raise typer.Abort()

    from protean.server import Engine

    engine = Engine(domain, test_mode=test_mode, debug=debug)
    engine.run()

    if engine.exit_code != 0:
        raise typer.Exit(code=engine.exit_code)



import inspect

from protean.container import Element, OptionsMixin
from protean.core.command import BaseCommand
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.utils import DomainObjects, derive_element_class, fully_qualified_name
from protean.utils.mixins import HandlerMixin


class BaseCommandHandler(Element, HandlerMixin, OptionsMixin):
    """Base Command Handler class that should implemented by all Domain CommandHandlers.

    This is also a marker class that is referenced when command handlers are registered
    with the domain
    """

    element_type = DomainObjects.COMMAND_HANDLER

    @classmethod
    def _default_options(cls):
        return [("part_of", None)]

    def __new__(cls, *args, **kwargs):
        if cls is BaseCommandHandler:
            raise NotSupportedError("BaseCommandHandler cannot be instantiated")
        return super().__new__(cls)


def command_handler_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseCommandHandler, **kwargs)

    if not element_cls.meta_.part_of:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Command Handler `{element_cls.__name__}` needs to be associated with an Aggregate"
                ]
            }
        )

    # Iterate through methods marked as `@handle` and construct a handler map
    if not element_cls._handlers:  # Protect against re-registration
        methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
        for method_name, method in methods:
            if not (
                method_name.startswith("__") and method_name.endswith("__")
            ) and hasattr(method, "_target_cls"):
                # Do not allow multiple handlers per command
                if (
                    fully_qualified_name(method._target_cls) in element_cls._handlers
                    and len(
                        element_cls._handlers[fully_qualified_name(method._target_cls)]
                    )
                    != 0
                ):
                    raise NotSupportedError(
                        f"Command {method._target_cls.__name__} cannot be handled by multiple handlers"
                    )

                # `_handlers` maps the command to its handler method
                element_cls._handlers[fully_qualified_name(method._target_cls)].add(
                    method
                )

                # Throw error if target_cls is not a Command
                if not inspect.isclass(method._target_cls) or not issubclass(
                    method._target_cls, BaseCommand
                ):
                    raise IncorrectUsageError(
                        {
                            "_command_handler": [
                                f"Method `{method_name}` in Command Handler `{element_cls.__name__}` "
                                "is not associated with a command"
                            ]
                        }
                    )

                # Associate Command with the handler's stream
                # Order of preference:
                #   1. Stream name defined in command
                #   2. Stream name derived from aggregate associated with command handler
                method._target_cls.meta_.stream_name = (
                    method._target_cls.meta_.stream_name
                    or element_cls.meta_.part_of.meta_.stream_name
                )

    return element_cls



import os
import re
import shutil
from typing import List, Optional

import typer
from copier import run_copy
from typing_extensions import Annotated

import protean


def new(
    project_name: Annotated[str, typer.Argument()],
    output_folder: Annotated[
        str, typer.Option("--output-dir", "-o", show_default=False)
    ] = ".",
    data: Annotated[List[str], typer.Option("--data", "-d", show_default=False)] = None,
    pretend: Annotated[Optional[bool], typer.Option("--pretend", "-p")] = False,
    force: Annotated[Optional[bool], typer.Option("--force", "-f")] = False,
    defaults: Annotated[Optional[bool], typer.Option("--defaults")] = False,
):
    def is_valid_project_name(project_name):
        """
        Validates the project name against criteria that ensure compatibility across
        Mac, Linux, and Windows systems, and also disallows spaces.
        """
        # Define a regex pattern that disallows the specified special characters
        # and spaces. This pattern also disallows leading and trailing spaces.
        forbidden_characters = re.compile(r'[<>:"/\\|?*\s]')

        if forbidden_characters.search(project_name) or not project_name:
            return False

        return True

    def clear_directory_contents(dir_path):
        """
        Removes all contents of a specified directory without deleting the directory itself.

        Parameters:
            dir_path (str): The path to the directory whose contents are to be cleared.
        """
        for item in os.listdir(dir_path):
            item_path = os.path.join(dir_path, item)
            if os.path.isfile(item_path) or os.path.islink(item_path):
                os.unlink(item_path)  # Remove files and links
            elif os.path.isdir(item_path):
                shutil.rmtree(item_path)  # Remove subdirectories and their contents

    if not is_valid_project_name(project_name):
        raise ValueError("Invalid project name")

    # Ensure the output folder exists
    if not os.path.isdir(output_folder):
        raise FileNotFoundError(f'Output folder "{output_folder}" does not exist')

    # The output folder is named after the project, and placed in the target folder
    project_directory = os.path.join(output_folder, project_name)

    # If the project folder already exists, and --force is not set, raise an error
    if os.path.isdir(project_directory) and os.listdir(project_directory):
        if not force:
            raise FileExistsError(
                f'Folder "{project_name}" is not empty. Use --force to overwrite.'
            )
        # Clear the directory contents if --force is set
        clear_directory_contents(project_directory)

    # Convert data tuples to a dictionary, if provided
    data_dict = {}
    for value in data:
        k, v = value.split("=", 1)
        data_dict[k] = v

    # Add the project name to answers
    data_dict["project_name"] = project_name

    # Create project from the cookiecutter-protean.git repo template
    run_copy(
        f"{protean.__path__[0]}/template",
        project_directory or ".",
        data=data_dict,
        unsafe=True,  # Trust our own template implicitly,
        defaults=defaults,
        pretend=pretend,
    )



from __future__ import annotations

import logging
import logging.config
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Dict, Union

from protean.core.command import BaseCommand
from protean.core.command_handler import BaseCommandHandler
from protean.core.event import BaseEvent
from protean.core.subscriber import BaseSubscriber
from protean.utils import DomainObjects, fully_qualified_name

logger = logging.getLogger(__name__)


class BaseBroker(metaclass=ABCMeta):
    """This class outlines the base broker functions,
    to be satisfied by all implementing brokers.

    It is also a marker interface for registering broker
    classes with the domain"""

    # FIXME Replace with typing.Protocol

    def __init__(
        self, name: str, domain: Any, conn_info: Dict[str, str]
    ) -> None:  # FIXME Any should be Domain
        self.name = name
        self.domain = domain
        self.conn_info = conn_info

        self._subscribers = defaultdict(set)
        self._command_handlers = {}

    @abstractmethod
    def publish(self, message: Dict) -> None:
        """Publish a message with Protean-compatible payload to the configured Message bus.

        Args:
            message (Dict): Command or Event payload
        """

    @abstractmethod
    def get_next(self) -> Dict:
        """Retrieve the next message to process from broker."""

    @abstractmethod
    def _data_reset(self) -> None:
        """Flush all data in broker instance.

        Useful for clearing cache and running tests.
        """

    def register(
        self,
        initiator_cls: Union[BaseCommand, BaseEvent],
        consumer_cls: Union[BaseCommandHandler, BaseSubscriber],
    ) -> None:
        """Registers Events and Commands with Subscribers/Command Handlers

        Arguments:
            initiator_cls {list} -- One or more Events or Commands
            consumer_cls {Subscriber/CommandHandler} -- The consumer class connected to the Event or Command
        """
        if not isinstance(initiator_cls, Iterable):
            initiator_cls = [initiator_cls]

        for initiator in initiator_cls:
            if initiator.element_type == DomainObjects.EVENT:
                self._subscribers[fully_qualified_name(initiator)].add(consumer_cls)
                logger.debug(
                    f"Registered Subscriber {consumer_cls.__name__} with broker {self.name}"
                )
            else:
                self._command_handlers[fully_qualified_name(initiator)] = consumer_cls



"""View Functionality and Classes"""

import logging

from protean.container import BaseContainer, OptionsMixin
from protean.core.entity import _EntityState
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.fields import Field, Reference, ValueObject
from protean.fields.association import Association
from protean.reflection import _ID_FIELD_NAME, declared_fields, id_field
from protean.utils import DomainObjects, derive_element_class, inflection

logger = logging.getLogger(__name__)


class BaseView(BaseContainer, OptionsMixin):
    element_type = DomainObjects.VIEW

    def __new__(cls, *args, **kwargs):
        if cls is BaseView:
            raise NotSupportedError("BaseView cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    def _default_options(cls):
        return [
            ("provider", "default"),
            ("cache", None),
            ("model", None),
            ("schema_name", inflection.underscore(cls.__name__)),
            ("order_by", ()),
        ]

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        if not subclass.meta_.abstract:
            subclass.__validate_id_field()

        subclass.__validate_for_basic_field_types()

    @classmethod
    def __validate_id_field(subclass):
        """Lookup the id field for this view and assign"""
        # FIXME What does it mean when there are no declared fields?
        #   Does it translate to an abstract view?
        if declared_fields(subclass):
            try:
                id_field = next(
                    field
                    for _, field in declared_fields(subclass).items()
                    if isinstance(field, (Field)) and field.identifier
                )

                setattr(subclass, _ID_FIELD_NAME, id_field.field_name)

            except StopIteration:
                # View does not have an ID field. An error will be thrown
                #   on registering the view, in the factory method.
                pass

    @classmethod
    def __validate_for_basic_field_types(subclass):
        for field_name, field_obj in declared_fields(subclass).items():
            if isinstance(field_obj, (Reference, Association, ValueObject)):
                raise IncorrectUsageError(
                    {
                        "_entity": [
                            f"Views can only contain basic field types. "
                            f"Remove {field_name} ({field_obj.__class__.__name__}) from class {subclass.__name__}"
                        ]
                    }
                )

    def __init__(self, *template, **kwargs):
        super().__init__(*template, **kwargs)

        # Set up the storage for instance state
        self.state_ = _EntityState()

    def __eq__(self, other):
        """Equivalence check to be based only on Identity"""

        # FIXME Enhanced Equality Checks
        #   * Ensure IDs have values and both of them are not null
        #   * Ensure that the ID is of the right type
        #   * Ensure that Objects belong to the same `type`
        #   * Check Reference equality

        # FIXME Check if `==` and `in` operator work with __eq__

        if type(other) is type(self):
            self_id = getattr(self, id_field(self).field_name)
            other_id = getattr(other, id_field(other).field_name)

            return self_id == other_id

        return False

    def __hash__(self):
        """Overrides the default implementation and bases hashing on identity"""

        # FIXME Add Object Class Type to hash
        return hash(getattr(self, id_field(self).field_name))


def view_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseView, **kwargs)

    if not element_cls.meta_.abstract and not hasattr(element_cls, _ID_FIELD_NAME):
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"View `{element_cls.__name__}` needs to have at least one identifier"
                ]
            }
        )

    element_cls.meta_.provider = (
        kwargs.pop("provider", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.provider)
        or "default"
    )
    element_cls.meta_.cache = (
        kwargs.pop("cache", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.cache)
        or None
    )
    element_cls.meta_.model = (
        kwargs.pop("model", None)
        or (hasattr(element_cls, "meta_") and element_cls.meta_.model)
        or None
    )

    if element_cls.meta_.provider and element_cls.meta_.cache:
        raise NotSupportedError(
            f"{element_cls.__name__} view can be persisted in"
            f"either a database or a cache, but not both"
        )

    return element_cls



from abc import abstractmethod

from protean import exceptions, utils
from protean.exceptions import ValidationError
from protean.globals import current_domain
from protean.reflection import association_fields, has_association_fields, id_field

from .base import Field, FieldBase
from .mixins import FieldCacheMixin, FieldDescriptorMixin


class _ReferenceField(Field):
    """
    Represents a reference field that can be used to establish associations between entities.

    Args:
        reference (str): The reference field as an attribute.
        **kwargs: Additional keyword arguments to be passed to the base `Field` class.
    """

    def __init__(self, reference, **kwargs):
        """Accept reference field as an attribute, otherwise is a straightforward field"""
        self.reference = reference
        super().__init__(**kwargs)

    def __set__(self, instance, value):
        """Override `__set__` to update relation field and keep it in sync with the shadow
        attribute's value

        Args:
            instance: The instance of the class.
            value: The value to be set.
        """
        value = self._load(value)

        if value:
            instance.__dict__[self.field_name] = value
        else:
            # Important to handle None assignment, and interpret it to mean resetting values
            self._reset_values(instance)

    def __delete__(self, instance):
        """Nullify values and linkages

        Args:
            instance: The instance of the class.
        """
        self._reset_values(instance)

    def _cast_to_type(self, value):
        """Verify the type of value assigned to the shadow field

        Args:
            value: The value to be assigned.

        Returns:
            The casted value.
        """
        # FIXME Verify that the value being assigned is compatible with the remote field
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self

        Args:
            value: The value to be converted to JSON.

        Raises:
            NotImplementedError: This method needs to be implemented in the derived class.

        """
        raise NotImplementedError

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items

        Args:
            instance: The instance of the class.
        """
        self.reference.value = None
        instance.__dict__.pop(self.field_name, None)
        instance.__dict__.pop(self.reference.field_name, None)
        self.reference.delete_cached_value(instance)


class Reference(FieldCacheMixin, Field):
    """
    A field representing a reference to another entity. This field is used to establish
    the reverse relationship to the remote entity.

    Args:
        to_cls (str or Entity): The target entity class or its name.
        **kwargs: Additional keyword arguments to be passed to the base `Field` class.
    """

    def __init__(self, to_cls, **kwargs):
        super().__init__(**kwargs)
        self._to_cls = to_cls

        self.relation = _ReferenceField(self)

    @property
    def to_cls(self):
        return self._to_cls

    def get_attribute_name(self):
        """Return formatted attribute name for the shadow field"""
        return "{}_{}".format(self.field_name, self.linked_attribute)

    def get_shadow_field(self):
        """Return shadow field
        Primarily used during Entity initialization to register shadow field"""
        return (self.attribute_name, self.relation)

    def get_cache_name(self):
        return self.field_name

    @property
    def linked_attribute(self):
        """Return linkage attribute to the target class

        This method is initially called from `__set_name__()` -> `get_attribute_name()`
        at which point, the `to_cls` has not been initialized properly. We simply default
        the linked attribute to 'id' in that case.

        Eventually, when setting value the first time, the `to_cls` entity is initialized
        and the attribute name is reset correctly.
        """
        if isinstance(self.to_cls, str):
            return "id"
        else:
            return id_field(self.to_cls).attribute_name

    def _resolve_to_cls(self, domain, to_cls, owner_cls):
        assert isinstance(self.to_cls, str)

        self._to_cls = to_cls

        # Refresh attribute name, now that we know `to_cls` Entity and it has been
        #   initialized with `id_field`
        self.attribute_name = self.get_attribute_name()

        # Reset the Shadow attribute's name
        self.relation = _ReferenceField(self)
        setattr(owner_cls, self.attribute_name, self.relation)
        self.relation.__set_name__(owner_cls, self.attribute_name)

        # Remove the earlier attribute if it is still attached
        old_attribute_name = "{}_{}".format(self.field_name, "id")
        if hasattr(owner_cls, old_attribute_name):
            delattr(owner_cls, old_attribute_name)

        # Update domain records because we enriched the class structure
        domain._replace_element_by_class(owner_cls)

    def __get__(self, instance, owner):
        """Retrieve associated objects"""
        reference_obj = None
        if hasattr(instance, "state_"):
            try:
                reference_obj = self.get_cached_value(instance)
            except KeyError:
                # Fetch target object by own Identifier
                id_field = self.get_attribute_name()

                id_value = None
                if hasattr(instance, id_field):
                    id_value = getattr(instance, id_field)

                if id_value:
                    reference_obj = self._fetch_objects(self.linked_attribute, id_value)
                    if reference_obj:
                        self._set_own_value(instance, reference_obj)
                    else:
                        # No Objects were found in the remote entity with this Entity's ID
                        pass

        return reference_obj

    def _fetch_objects(self, key, value):
        """Fetch Multiple linked objects"""
        return current_domain.repository_for(self.to_cls)._dao.find_by(**{key: value})

    def __set__(self, instance, value):
        """Override `__set__` to coordinate between relation field and its shadow attribute"""
        value = self._load(value)

        if value:
            # Check if the reference object has been saved. Otherwise, throw ValueError
            # FIXME not a comprehensive check. Should refer to state
            if getattr(value, id_field(value).field_name) is None:
                raise ValueError(
                    "Target Object must be saved before being referenced",
                    self.field_name,
                )
            else:
                self._set_own_value(instance, value)
                self._set_relation_value(
                    instance, getattr(value, self.linked_attribute)
                )
        else:
            self._reset_values(instance)

    def _set_own_value(self, instance, value):
        if value is None:
            instance.__dict__.pop(self.field_name, None)
            self.delete_cached_value(instance)
        else:
            instance.__dict__[self.field_name] = value
            self.set_cached_value(instance, value)

        # Mark Entity as Dirty
        if hasattr(instance, "state_"):
            instance.state_.mark_changed()

    def _set_relation_value(self, instance, value):
        if value is None:
            instance.__dict__.pop(self.attribute_name, None)
        else:
            instance.__dict__[self.attribute_name] = value

    def __delete__(self, instance):
        self._reset_values(instance)

    def _reset_values(self, instance):
        """Reset all associated values and clean up dictionary items"""
        self._set_own_value(instance, None)
        self._set_relation_value(instance, None)

    def _cast_to_type(self, value):
        # FIXME Assign value only of the correct type
        # if not isinstance(value, self.to_cls):
        #     self.fail('invalid', value=value)
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        raise NotImplementedError


class Association(FieldBase, FieldDescriptorMixin, FieldCacheMixin):
    """
    Represents an association between entities in a domain model.

    An association field allows one entity to reference another entity in the domain model.
    It provides methods to retrieve associated objects and handle changes in the association.

    Args:
        to_cls (class): The class of the target entity that this association references.
    """

    def __init__(self, to_cls, **kwargs):
        super().__init__(**kwargs)

        self._to_cls = to_cls

        # FIXME Find an elegant way to avoid these declarations in associations
        # Associations cannot be marked `required` or `unique`
        self.required = False
        self.unique = False

    @property
    def to_cls(self):
        return self._to_cls

    def _resolve_to_cls(self, domain, to_cls, owner_cls):
        """Resolves class references to actual class object.

        Called by the domain when a new element is registered,
        and its name matches `to_cls`
        """
        self._to_cls = to_cls

    def _cast_to_type(self, value):
        """Verify type of value assigned to the association field"""
        # FIXME Verify that the value being assigned is compatible with the associated Entity
        return value

    def _linked_attribute(self, owner):
        """Return linkage attribute to own entity's `id_field`

        FIXME Explore converting this method into an attribute, and treating it
        uniformly at `association` level.
        """
        return (
            utils.inflection.underscore(owner.__name__)
            + "_"
            + id_field(owner).attribute_name
        )

    def _linked_reference(self, owner):
        return utils.inflection.underscore(owner.__name__)

    def __get__(self, instance, owner):
        """Retrieve associated objects"""

        try:
            reference_obj = self.get_cached_value(instance)
        except KeyError:
            # Fetch target object by own Identifier
            id_value = getattr(instance, id_field(instance).field_name)
            reference_obj = self._fetch_objects(
                instance, self._linked_attribute(owner), id_value
            )

            self._set_own_value(instance, reference_obj)

        return reference_obj

    def _set_own_value(self, instance, value):
        instance.__dict__[self.field_name] = value
        self.set_cached_value(instance, value)

        # Mark Entity as Dirty
        if hasattr(instance, "state_"):
            instance.state_.mark_changed()

    @abstractmethod
    def _fetch_objects(self, instance, key, value):
        """Placeholder method for customized Association query methods"""
        raise NotImplementedError

    @abstractmethod
    def as_dict(self):
        """Return JSON-compatible value of field"""
        raise NotImplementedError

    def __set__(self, instance, value):
        """Set the value of the association field"""
        # Preserve heirarchy of entities.
        #
        #   Owner: is the entity that owns the association field
        #   Root: is the entity that is at the top of the hierarchy, an Aggregate Root
        if value is not None:
            items = value if isinstance(value, list) else [value]
            for item in items:
                item._set_root_and_owner(instance._root, instance)

    def __delete__(self, instance):
        """Cannot pop values for an association"""
        raise exceptions.NotSupportedError(
            "Object does not support the operation being performed",
            self.field_name,
        )

    def get_cache_name(self):
        return self.field_name

    @property
    def has_changed(self):
        return self.change is not None


class HasOne(Association):
    """
    Represents an one-to-one association between an aggregate and its entities.

    This field is used to define a relationship where an aggregate is associated
    with at most one instance of a child entity.
    """

    def __set__(self, instance, value):
        """Setup relationship to be persisted/updated

        We track the change in the instance's `_temp_cache` to determine if the relationship
        has been added, updated, or deleted. We track two aspects: state and old value.

        For `HasOne`, there are three possible states:
        - ADDED: The relationship is being added for the first time
        - UPDATED: The relationship is being updated
        - DELETED: The relationship is being removed

        Of these, the old value is applicable for `UPDATED` and `DELETED` states.

        Also, we recursively remove child entities if they are associated with the old value.

        The `temp_cache` we set up here is eventually used by the `Repository` to determine
        the changes to be persisted.
        """

        super().__set__(instance, value)

        if value is not None and not isinstance(value, self.to_cls):
            raise ValidationError(
                {
                    "_entity": [
                        f"Value assigned to '{self.field_name}' is not of type '{self.to_cls.__name__}'"
                    ]
                }
            )

        # 1. Preserve parent linkage in child entity
        if value is not None:
            # This updates the parent's unique identifier in the child
            #   so that the foreign key relationship is preserved
            id_value = getattr(instance, id_field(instance).field_name)
            linked_attribute = self._linked_attribute(instance.__class__)
            if hasattr(value, linked_attribute):
                setattr(
                    value, linked_attribute, id_value
                )  # This overwrites any existing linkage, which is correct

            # Add the parent to the child entity cache
            # Temporarily set linkage to parent in child entity
            setattr(value, self._linked_reference(type(instance)), instance)

        # 2. Determine and store the change in the relationship
        current_value = getattr(instance, self.field_name)
        if current_value is None:
            # Entity was not associated earlier
            instance._temp_cache[self.field_name]["change"] = "ADDED"
        elif value is None:
            # Entity was associated earlier, but now being removed
            instance._temp_cache[self.field_name]["change"] = "DELETED"
            instance._temp_cache[self.field_name]["old_value"] = current_value
        elif current_value.id != value.id:
            # A New Entity is being associated replacing the old one
            instance._temp_cache[self.field_name]["change"] = "UPDATED"
            instance._temp_cache[self.field_name]["old_value"] = current_value
        elif current_value.id == value.id and value.state_.is_changed:
            # Entity was associated earlier, but now being updated
            instance._temp_cache[self.field_name]["change"] = "UPDATED"
        else:
            instance._temp_cache[self.field_name]["change"] = (
                None  # The same object has been assigned, No-Op
            )

        self._set_own_value(instance, value)

        # 3. Go Recursive and remove child entities if they are associated with the old value
        if instance._temp_cache[self.field_name]["change"] == "DELETED":
            old_value = instance._temp_cache[self.field_name]["old_value"]
            if has_association_fields(old_value):
                for field_name, field_obj in association_fields(old_value).items():
                    if isinstance(field_obj, HasMany):
                        field_obj.remove(old_value, getattr(old_value, field_name))
                    elif isinstance(field_obj, HasOne):
                        setattr(old_value, field_name, None)

        if instance._initialized and instance._root is not None:
            instance._root._postcheck()  # Trigger validations from the top

    def _fetch_objects(self, instance, key, identifier):
        """Fetch single linked object"""
        try:
            repo = current_domain.repository_for(self.to_cls)
            value = repo._dao.find_by(**{key: identifier})

            # Set up linkage with owner element
            setattr(
                value, key, identifier
            )  # This overwrites any existing linkage, which is correct

            return value
        except exceptions.ObjectNotFoundError:
            return None

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        if value is not None:
            return value.to_dict()


class HasMany(Association):
    """
    Represents a one-to-many association between two entities. This field is used to define a relationship where an
    aggregate has multiple instances of a chil entity.

    Args:
        to_cls (class): The class of the target entity.
        **kwargs: Additional keyword arguments to be passed to the base field class.
    """

    def __set__(self, instance, value):
        """This supports direct assignment of values to HasMany fields, like:
        `order.items = [item1, item2, item3]`
        """
        super().__set__(instance, value)

        if value is not None:
            self.add(instance, value)

    def add(self, instance, items) -> None:
        """
        Available as `add_<HasMany Field Name>` method on the entity instance.

        This method adds one or more linked entities to the source entity. It also diffs the current value with the
        new value to determine the changes that need to be persisted.

        The method also takes care of the attributes of the linked entities, preparing them for persistence.

        Each linkage takes care of its own attributes, preparing them for persistence.
        One exception is deletion of an entity - all child entities have to be marked as removed.

        We track the change in the instance's `_temp_cache` to determine if the relationship
        has been added, updated, or deleted. We track each item's state and group the changes
        into three buckets:
        - ADDED: The relationship is being added for the first time
        - UPDATED: The relationship is being updated
        - DELETED: The relationship is being removed

        The `DELETED` objects are detected from the pool of new objects, but it is also possible to remove them
        directly with the `remove` method.

        The `temp_cache` we set up here is eventually used by the `Repository` to determine
        the changes to be persisted.

        Args:
            instance: The source entity instance.
            items: The linked entity or entities to be added.
        """
        super().__set__(instance, items)

        data = getattr(instance, self.field_name)

        # Convert a single item into a list of items, if necessary
        items = [items] if not isinstance(items, list) else items

        # Validate that all items are of the same type, and the correct type
        for item in items:
            if not isinstance(item, self.to_cls):
                raise ValidationError(
                    {
                        "_entity": [
                            f"Value assigned to '{self.field_name}' is not of type '{self.to_cls.__name__}'"
                        ]
                    }
                )

        current_value_ids = [value.id for value in data]

        # Remove items when set to empty
        if len(items) == 0 and len(current_value_ids) > 0:
            self.remove(instance, data)

        for item in items:
            # Items to add
            if item.id not in current_value_ids:
                # If the same item is added multiple times, the last item added will win
                instance._temp_cache[self.field_name]["added"][item.id] = item

                setattr(
                    item,
                    self._linked_attribute(type(instance)),
                    getattr(instance, id_field(instance).field_name),
                )

                # Temporarily set linkage to parent in child entity
                setattr(item, self._linked_reference(type(instance)), instance)

                # Reset Cache
                self.delete_cached_value(instance)
            # Items to update
            elif (
                item.id in current_value_ids
                and item.state_.is_persisted
                and item.state_.is_changed
            ):
                setattr(
                    item,
                    self._linked_attribute(type(instance)),
                    getattr(instance, id_field(instance).field_name),
                )

                # Temporarily set linkage to parent in child entity
                setattr(item, self._linked_reference(type(instance)), instance)

                instance._temp_cache[self.field_name]["updated"][item.id] = item

                # Reset Cache
                self.delete_cached_value(instance)

        if instance._initialized and instance._root is not None:
            instance._root._postcheck()  # Trigger validations from the top

    def remove(self, instance, items) -> None:
        """
        Available as `add_<HasMany Field Name>` method on the entity instance.

        Remove one or more linked entities from the source entity.

        We also recursively remove child entities if they are associated with the removed value.

        Args:
            instance: The source entity instance.
            items: The linked entity or entities to be removed.
        """
        data = getattr(instance, self.field_name)

        # Convert a single item into a list of items, if necessary
        items = [items] if not isinstance(items, list) else items

        # Validate that all items are of the same type, and the correct type
        for item in items:
            if not isinstance(item, self.to_cls):
                raise ValidationError(
                    {
                        "_entity": [
                            f"Value assigned to '{self.field_name}' is not of type '{self.to_cls.__name__}'"
                        ]
                    }
                )

        current_value_ids = [value.id for value in data]

        for item in items:
            if item.id in current_value_ids:
                if item.id not in instance._temp_cache[self.field_name]["removed"]:
                    instance._temp_cache[self.field_name]["removed"][item.id] = item

                    # Reset Cache
                    self.delete_cached_value(instance)

            # Remove child entities
            if has_association_fields(item):
                for field_name, field_obj in association_fields(item).items():
                    if isinstance(field_obj, HasMany):
                        field_obj.remove(item, getattr(item, field_name))
                    elif isinstance(field_obj, HasOne):
                        setattr(item, field_name, None)

        if instance._initialized and instance._root is not None:
            instance._root._postcheck()  # Trigger validations from the top

    def _fetch_objects(self, instance, key, value) -> list:
        """
        Fetch linked entities.

        Args:
            instance: The source entity instance.
            key (str): The name of the attribute on the target entity that links back to the source entity.
            value: The value of the foreign key.

        Returns:
            list: A list of linked entity instances.
        """
        children_repo = current_domain.repository_for(self.to_cls)
        data = children_repo._dao.query.filter(**{key: value}).all().items

        # Set up linkage with owner element
        for item in data:
            setattr(item, key, value)

        # Add objects in temporary cache
        for _, item in instance._temp_cache[self.field_name]["added"].items():
            data.append(item)

        # Update objects from temporary cache if present
        updated_objects = []
        for value in data:
            if value.id in instance._temp_cache[self.field_name]["updated"]:
                updated_objects.append(
                    instance._temp_cache[self.field_name]["updated"][value.id]
                )
            else:
                updated_objects.append(value)
        data = updated_objects

        # Remove objects marked as removed in temporary cache
        for _, item in instance._temp_cache[self.field_name]["removed"].items():
            data[:] = [value for value in data if value.id != item.id]

        return data

    def as_dict(self, value) -> list:
        """
        Return JSON-compatible value of self.

        Args:
            value: The value to be converted to a JSON-compatible format.

        Returns:
            list: A list of dictionaries representing the linked entities.
        """
        return [item.to_dict() for item in value]

    def get(self, instance, **kwargs):
        """Fetch a single linked entity based on the provided criteria.

        Available as `get_one_from_<HasMany Field Name>` method on the entity instance.

        Args:
            **kwargs: The filtering criteria.
        """
        data = self.filter(instance, **kwargs)

        if len(data) == 0:
            raise exceptions.ObjectNotFoundError(
                {"self.field_name": ["No linked entities matching criteria found"]}
            )

        if len(data) > 1:
            raise exceptions.TooManyObjectsError(
                {
                    "self.field_name": [
                        "Multiple linked entities matching criteria found"
                    ]
                }
            )

        return data[0]

    def filter(self, instance, **kwargs):
        """Filter the linked entities based on the provided criteria.

        Available as `filter_<HasMany Field Name>` method on the entity instance.

        Args:
            **kwargs: The filtering criteria.
        """
        data = getattr(instance, self.field_name)
        return [
            item
            for item in data
            if all(getattr(item, key) == value for key, value in kwargs.items())
        ]



import logging

from protean.exceptions import (
    ExpectedVersionError,
    InvalidOperationError,
    ValidationError,
)
from protean.globals import _uow_context_stack, current_domain
from protean.reflection import id_field
from protean.utils import DomainObjects, EventProcessing, fqn

logger = logging.getLogger(__name__)


class UnitOfWork:
    def __init__(self):
        """Initialize session factories from all providers

        Connections will be retrieved at this stage

        Also initialize Identity Map?
        Repository will first check here before retrieving from Database
        """
        # FIXME Should UnitOfWork keep an Identity map, of all `seen` objects?
        self.domain = current_domain
        self._in_progress = False

        self._sessions = {}
        self._messages_to_dispatch = []
        self._identity_map = {}

    @property
    def in_progress(self):
        return self._in_progress

    def __enter__(self):
        # Initiate a new session as part of self
        self.start()
        return self

    def __exit__(self, *args):
        # Commit and destroy session
        self.commit()

    def _add_to_identity_map(self, aggregate) -> None:
        identifier = getattr(aggregate, id_field(aggregate).field_name)
        self._identity_map[identifier] = aggregate

    def start(self):
        # Stand in method for `__enter__`
        #   To explicitly begin and end transactions
        self._in_progress = True
        _uow_context_stack.push(self)

    def commit(self):  # noqa: C901
        # Raise error if there the Unit Of Work is not active
        logger.debug(f"Committing {self}...")
        if not self._in_progress:
            raise InvalidOperationError("UnitOfWork is not in progress")

        # Exit from Unit of Work
        _uow_context_stack.pop()

        # Commit and destroy session
        try:
            for _, session in self._sessions.items():
                session.commit()

            # Push messages to all brokers
            # FIXME Send message to its designated broker?
            # FIXME Send messages through domain.brokers.publish?
            for message in self._messages_to_dispatch:
                for _, broker in self.domain.brokers.items():
                    broker.publish(message)
            self._messages_to_dispatch = []  # Empty after dispatch

            events = []
            for _, item in self._identity_map.items():
                if item._events:
                    if item.element_type == DomainObjects.EVENT_SOURCED_AGGREGATE:
                        for event in item._events:
                            current_domain.event_store.store.append_aggregate_event(
                                item, event
                            )
                            events.append((item, event))
                    else:
                        for event in item._events:
                            current_domain.event_store.store.append(event)
                            events.append((item, event))
                item._events = []

            # Iteratively consume all events produced in this session
            if current_domain.config["event_processing"] == EventProcessing.SYNC.value:
                # Handover events to process instantly
                for _, event in events:
                    handler_classes = current_domain.handlers_for(event)
                    for handler_cls in handler_classes:
                        handler_methods = (
                            handler_cls._handlers[fqn(event.__class__)]
                            or handler_cls._handlers["$any"]
                        )

                        for handler_method in handler_methods:
                            handler_method(handler_cls(), event)

            logger.debug("Commit Successful")
        except ValueError as exc:
            logger.error(str(exc))
            self.rollback()

            # Extact message based on message store platform in use
            if str(exc).startswith("P0001-ERROR"):
                msg = str(exc).split("P0001-ERROR:  ")[1]
            else:
                msg = str(exc)
            raise ExpectedVersionError(msg) from None
        except Exception as exc:
            logger.error(
                f"Error during Commit: {str(exc)}. Rolling back Transaction..."
            )
            self.rollback()
            raise ValidationError(
                {"_entity": [f"Error during Data Commit: - {repr(exc)}"]}
            )

        self._reset()

    def _reset(self):
        for _, session in self._sessions.items():
            session.close()

        self._sessions = {}
        self._messages_to_dispatch = []
        self._identity_map = set()
        self._in_progress = False

    def rollback(self):
        # Raise error if the Unit Of Work is not active
        if not self._in_progress:
            raise InvalidOperationError("UnitOfWork is not in progress")

        # Exit from Unit of Work
        _uow_context_stack.pop()

        try:
            for _, session in self._sessions.items():
                session.rollback()

            logger.debug("Transaction rolled back")
        except Exception as exc:
            logger.error(f"Error during Transaction rollback: {str(exc)}")

        self._reset()

    def _get_session(self, provider_name):
        provider = self.domain.providers[provider_name]
        return provider.get_session()

    def _initialize_session(self, provider_name):
        new_session = self._get_session(provider_name)
        self._sessions[provider_name] = new_session
        if not new_session.is_active:
            new_session.begin()
        return new_session

    def get_session(self, provider_name):
        if provider_name in self._sessions:
            return self._sessions[provider_name]
        else:
            return self._initialize_session(provider_name)

    def register_message(self, message):  # FIXME Add annotations
        self._messages_to_dispatch.append(message)



import logging
from abc import abstractmethod
from typing import Any, Optional

from protean.container import Element, OptionsMixin
from protean.core.event import BaseEvent
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger(__name__)


class BaseSubscriber(Element, OptionsMixin):
    """Base Subscriber class that should implemented by all Domain Subscribers.

    This is also a marker class that is referenced when subscribers are registered
    with the domain
    """

    element_type = DomainObjects.SUBSCRIBER

    def __new__(cls, *args, **kwargs):
        if cls is BaseSubscriber:
            raise NotSupportedError("BaseSubscriber cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    def _default_options(cls):
        return [("event", None), ("broker", "default")]

    @abstractmethod
    def __call__(self, event: BaseEvent) -> Optional[Any]:
        """Placeholder method for receiving notifications on event"""
        raise NotImplementedError


def subscriber_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseSubscriber, **kwargs)

    if not element_cls.meta_.event:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Subscriber `{element_cls.__name__}` needs to be associated with an Event"
                ]
            }
        )

    if not element_cls.meta_.broker:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Subscriber `{element_cls.__name__}` needs to be associated with a Broker"
                ]
            }
        )

    return element_cls



from typing import Any

from protean.exceptions import IncorrectUsageError

_FIELDS = "__container_fields__"
_ID_FIELD_NAME = "__container_id_field_name__"


def fields(class_or_instance):
    """Return a tuple describing the fields of this dataclass.

    Accepts a dataclass or an instance of one. Tuple elements are of
    type Field.
    """

    # Might it be worth caching this, per class?
    try:
        fields_dict = getattr(class_or_instance, _FIELDS)
    except AttributeError:
        raise IncorrectUsageError(
            {"field": [f"{class_or_instance} does not have fields"]}
        )

    return fields_dict


def id_field(class_or_instance):
    try:
        field_name = getattr(class_or_instance, _ID_FIELD_NAME)
    except AttributeError:
        return None

    return fields(class_or_instance)[field_name]


def has_id_field(class_or_instance: Any) -> bool:
    """Check if class/instance has an identity attribute.

    Args:
        class_or_instance (Any): Domain Element to check.

    Returns:
        bool: True if the element has an identity field.
    """
    return hasattr(class_or_instance, _ID_FIELD_NAME)


def has_fields(class_or_instance):
    """Check if Protean element encloses fields"""
    return hasattr(class_or_instance, _FIELDS)


def attributes(class_or_instance):
    attributes_dict = {}

    for _, field_obj in fields(class_or_instance).items():
        # FIXME Make these checks elegant
        # Because of circular import issues, `Reference` class cannot be imported
        #   in this file. So we are resorting to check for method presence in
        #   field objects. Not the most elegant way, but will have to suffice
        #   until class heirarchies are restructured.
        if hasattr(field_obj, "get_shadow_fields"):
            shadow_fields = field_obj.get_shadow_fields()
            for _, shadow_field in shadow_fields:
                attributes_dict[shadow_field.attribute_name] = shadow_field
        elif hasattr(field_obj, "relation"):
            attributes_dict[field_obj.get_attribute_name()] = field_obj.relation
        elif not hasattr(field_obj, "to_cls"):
            attributes_dict[field_obj.get_attribute_name()] = field_obj
        else:  # This field is an association. Ignore recording it as an attribute
            pass

    return attributes_dict


def unique_fields(class_or_instance):
    """Return fields marked as unique for this class or instance"""
    return {
        field_name: field_obj
        for field_name, field_obj in attributes(class_or_instance).items()
        if field_obj.unique
    }


def declared_fields(class_or_instance):
    """Return a tuple describing the declared fields of this dataclass.

    Accepts a dataclass or an instance of one. Tuple elements are of
    type Field.

    `_version` is a auto-controlled, internal field, so is not returned
    among declared fields.
    """

    # Might it be worth caching this, per class?
    try:
        fields_dict = dict(getattr(class_or_instance, _FIELDS))
        fields_dict.pop("_version", None)
    except AttributeError:
        raise IncorrectUsageError(
            {"field": [f"{class_or_instance} does not have fields"]}
        )

    return fields_dict


def association_fields(class_or_instance):
    """Return a tuple describing the association fields of this dataclass.

    Accepts an Entity. Tuple elements are of type Field.
    """
    from protean.fields.association import Association

    return {
        field_name: field_obj
        for field_name, field_obj in declared_fields(class_or_instance).items()
        if isinstance(field_obj, Association)
    }


def has_association_fields(class_or_instance):
    """Check if Protean element encloses association fields"""
    return bool(association_fields(class_or_instance))



from typing import TYPE_CHECKING, Dict

from protean.port.broker import BaseBroker
from protean.utils import fully_qualified_name
from protean.utils.mixins import Message

if TYPE_CHECKING:
    from protean.domain import Domain


class InlineBroker(BaseBroker):
    def __init__(self, name: str, domain: "Domain", conn_info: Dict[str, str]) -> None:
        super().__init__(name, domain, conn_info)

        # In case of `InlineBroker`, the `IS_ASYNC` value will always be `False`.
        conn_info["IS_ASYNC"] = False

    def publish(self, message: Message) -> None:
        initiator_obj = message.to_object()

        for subscriber in self._subscribers[
            fully_qualified_name(initiator_obj.__class__)
        ]:
            subscriber()(message.data)

    def get_next(self) -> Dict:
        """No-Op"""
        return None

    def _data_reset(self) -> None:
        """No-Op"""



import collections
import importlib
import logging

from protean.core.event import BaseEvent
from protean.exceptions import ConfigurationError
from protean.globals import current_uow
from protean.utils.mixins import Message

logger = logging.getLogger(__name__)


BROKER_PROVIDERS = {
    "inline": "protean.adapters.InlineBroker",
    "redis": "protean.adapters.broker.redis.RedisBroker",
    "celery": "protean.adapters.broker.celery.CeleryBroker",
}


class Brokers(collections.abc.MutableMapping):
    def __init__(self, domain):
        self.domain = domain
        self._brokers = None

    def __getitem__(self, key):
        return self._brokers[key] if self._brokers else None

    def __iter__(self):
        return iter(self._brokers) if self._brokers else iter({})

    def __len__(self):
        return len(self._brokers) if self._brokers else 0

    def __setitem__(self, key, value):
        if self._brokers is None:
            self._brokers = {}

        self._brokers[key] = value

    def __delitem__(self, key):
        if key in self._brokers:
            del self._brokers[key]

    def _initialize(self):
        """Read config file and initialize brokers"""
        configured_brokers = self.domain.config["brokers"]
        broker_objects = {}

        logger.debug("Initializing brokers...")
        if configured_brokers and isinstance(configured_brokers, dict):
            if "default" not in configured_brokers:
                raise ConfigurationError("You must define a 'default' broker")

            for broker_name, conn_info in configured_brokers.items():
                broker_full_path = BROKER_PROVIDERS[conn_info["provider"]]
                broker_module, broker_class = broker_full_path.rsplit(".", maxsplit=1)

                broker_cls = getattr(
                    importlib.import_module(broker_module), broker_class
                )
                broker_objects[broker_name] = broker_cls(broker_name, self, conn_info)
        else:
            raise ConfigurationError("Configure at least one broker in the domain")

        self._brokers = broker_objects

        # Initialize subscribers for Brokers
        for _, subscriber_record in self.domain.registry.subscribers.items():
            subscriber = subscriber_record.cls
            broker_name = subscriber.meta_.broker

            if broker_name not in self._brokers:
                raise ConfigurationError(
                    f"Broker `{broker_name}` has not been configured."
                )

            self._brokers[broker_name].register(subscriber.meta_.event, subscriber)

    def publish(self, object: BaseEvent) -> None:
        """Publish an object to all registered brokers"""
        message = Message.to_message(object)

        # Follow a naive strategy and dispatch event directly to message broker
        #   If the operation is enclosed in a Unit of Work, delegate the responsibility
        #   of publishing the message to the UoW
        if current_uow:
            logger.debug(
                f"Recording {object.__class__.__name__} "
                f"with values {object.to_dict()} in {current_uow}"
            )
            current_uow.register_message(message)
        else:
            logger.debug(
                f"Publishing {object.__class__.__name__} with values {object.to_dict()}"
            )
            for _, broker in self._brokers.items():
                broker.publish(message)



import json
from typing import TYPE_CHECKING, Dict

import redis

from protean.port.broker import BaseBroker
from protean.utils.mixins import Message

if TYPE_CHECKING:
    from protean.domain import Domain


class RedisBroker(BaseBroker):
    """Redis as the Message Broker.

    FIXME: Convert to be a Context Manager, and release connection after use
    """

    def __init__(self, name: str, domain: "Domain", conn_info: Dict) -> None:
        super().__init__(name, domain, conn_info)

        self.redis_instance = redis.Redis.from_url(conn_info["URI"])

    def publish(self, message: Message) -> None:
        # FIXME Accept configuration for database and list name
        self.redis_instance.rpush("messages", json.dumps(message.to_dict()))

    def get_next(self) -> Message:
        bytes_message = self.redis_instance.lpop("messages")
        if bytes_message:
            return Message(json.loads(bytes_message))

        return None

    def _data_reset(self) -> None:
        self.redis_instance.flushall()



from abc import ABCMeta, abstractmethod
from collections import deque
from typing import Any, Dict, List, Optional, Type, Union

from protean import BaseCommand, BaseEvent, BaseEventSourcedAggregate
from protean.fields import Identifier
from protean.utils.mixins import Message


class BaseEventStore(metaclass=ABCMeta):
    """This class outlines the base event store capabilities
    to be implemented in all supported event store adapters.

    It is also a marker interface for registering event store
    classes with the domain.
    """

    def __init__(
        self, name: str, domain: Any, conn_info: Dict[str, str]
    ) -> None:  # FIXME Any should be Domain
        self.name = name
        self.domain = domain
        self.conn_info = conn_info

    @abstractmethod
    def _write(
        self,
        stream_name: str,
        message_type: str,
        data: Dict,
        metadata: Dict | None = None,
        expected_version: int | None = None,
    ) -> int:
        """Write a message to the event store.

        Returns the position of the message in the stream.

        Implemented by the concrete event store adapter.
        """

    @abstractmethod
    def _read(
        self,
        stream_name: str,
        sql: str | None = None,
        position: int = 0,
        no_of_messages: int = 1000,
    ) -> List[Dict[str, Any]]:
        """Read messages from the event store.

        Implemented by the concrete event store adapter.
        """

    @abstractmethod
    def _read_last_message(self, stream_name) -> Dict[str, Any]:
        """Read the last message from the event store.

        Implemented by the concrete event store adapter.
        """

    def category(self, stream_name: str) -> str:
        if not stream_name:
            return ""

        stream_category, _, _ = stream_name.partition("-")
        return stream_category

    def read(
        self,
        stream_name: str,
        sql: str | None = None,
        position: int = 0,
        no_of_messages: int = 1000,
    ):
        raw_messages = self._read(
            stream_name, sql=sql, position=position, no_of_messages=no_of_messages
        )

        messages = []
        for raw_message in raw_messages:
            messages.append(Message.from_dict(raw_message))

        return messages

    def read_last_message(self, stream_name) -> Message:
        # FIXME Rename to read_last_stream_message
        raw_message = self._read_last_message(stream_name)
        return Message.from_dict(raw_message)

    def append_aggregate_event(
        self, aggregate: BaseEventSourcedAggregate, event: BaseEvent
    ) -> int:
        message = Message.to_aggregate_event_message(aggregate, event)

        position = self._write(
            message.stream_name,
            message.type,
            message.data,
            metadata=message.metadata.to_dict(),
            expected_version=message.expected_version,
        )

        # Increment aggregate's version as we process events
        #    to correctly handle expected version
        aggregate._version += 1

        return position

    def append(self, object: Union[BaseEvent, BaseCommand]) -> int:
        message = Message.to_message(object)

        return self._write(
            message.stream_name,
            message.type,
            message.data,
            metadata=message.metadata.to_dict(),
        )

    def load_aggregate(
        self, part_of: Type[BaseEventSourcedAggregate], identifier: Identifier
    ) -> Optional[BaseEventSourcedAggregate]:
        """Load an aggregate from underlying events.

        The first event is used to initialize the aggregate, after which each event is
        applied in sequence until the end.

        A snapshot, if one exists, is loaded first and subsequent events are
        applied on it. If there are more than SNAPSHOT_THRESHOLD events since snapshot,
        a new snapshot is written to the store.

        Args:
            part_of (Type[BaseEventSourcedAggregate]): The EventSourced Aggregate's class
            identifier (Identifier): Unique aggregate identifier

        Returns:
            Optional[BaseEventSourcedAggregate]: Return fully-formed aggregate when events exist,
                or None.
        """
        snapshot = self._read_last_message(
            f"{part_of.meta_.stream_name}:snapshot-{identifier}"
        )

        if snapshot:
            aggregate = part_of(**snapshot["data"])
            position_in_snapshot = snapshot["data"]["_version"]

            events = deque(
                self._read(
                    f"{part_of.meta_.stream_name}-{identifier}",
                    position=position_in_snapshot + 1,
                )
            )
        else:
            events = deque(self._read(f"{part_of.meta_.stream_name}-{identifier}"))

            if not events:
                return None

            # Handle first event separately to initialize the aggregate
            first_event = events.popleft()
            aggregate = part_of(**first_event["data"])

            # Also apply the first event in case a method has been specified
            aggregate._apply(first_event)
            aggregate._version += 1

        # Apply all other events one-by-one
        for event in events:
            aggregate._apply(event)
            aggregate._version += 1

        # If there are more events than SNAPSHOT_THRESHOLD, create a new snapshot
        if (
            snapshot
            and len(events) > 1
            and (
                events[-1]["position"] - position_in_snapshot
                >= self.domain.config["SNAPSHOT_THRESHOLD"]
            )
        ) or (
            not snapshot and len(events) + 1 >= self.domain.config["SNAPSHOT_THRESHOLD"]
        ):  # Account for the first event that was popped
            self._write(
                f"{part_of.meta_.stream_name}:snapshot-{identifier}",
                "SNAPSHOT",
                aggregate.to_dict(),
            )

        return aggregate

    @abstractmethod
    def _data_reset(self) -> None:
        """Flush all events.

        Useful for running tests with a clean slate.
        """
